"""
Action selection strategies
===========================

These selection strategies implement a variety of policies for choosing which focused
transmit to fire next, potentially given some beliefs about the state of tissue,
as represented by ``particles``.

For a comprehensive example usage, see: :doc:`../notebooks/agent/agent_example`

All strategies are stateless, meaning that they do not maintain any internal state.
"""

import keras
from keras import ops

from zea import tensor_ops
from zea.agent import masks
from zea.internal.registry import action_selection_registry


class MaskActionModel:
    """Base class for any action selection method that does masking."""

    def apply(self, action, observation):
        """Apply the action to the observation.

        Args:
            action (Tensor): The mask to be applied.
            observation (Tensor): The observation to which the action is applied.

        Returns:
            Tensor: The masked tensor
        """
        return observation * action


class LinesActionModel(MaskActionModel):
    """Base class for action selection methods that select lines."""

    def __init__(self, n_actions: int, n_possible_actions: int, img_width: int, img_height: int):
        """Initialize the LinesActionModel.

        Args:
            n_actions (int): The number of actions the agent can take.
            n_possible_actions (int): The number of possible actions.
            img_width (int): The width of the input image.
            img_height (int): The height of the input image.

        Attributes:
            stack_n_cols (int): The number of columns in the image that correspond
                to a single action.
        """
        super().__init__()
        self.n_actions = n_actions
        self.n_possible_actions = n_possible_actions
        self.img_width = img_width
        self.img_height = img_height

        stack_n_cols = self.img_width / self.n_possible_actions
        assert stack_n_cols.is_integer(), "Image width must be divisible by n_possible_actions."
        self.stack_n_cols = int(stack_n_cols)

    def lines_to_im_size(self, lines):
        """Convert k-hot-encoded line vectors to image size.

        Args:
            lines (Tensor): shape is (n_masks, n_possible_actions)

        Returns:
            Tensor: Masks of shape (n_masks, img_height, img_width)
        """
        return masks.lines_to_im_size(lines, (self.img_height, self.img_width))


@action_selection_registry(name="greedy_entropy")
class GreedyEntropy(LinesActionModel):
    """Greedy entropy action selection.

    Selects the max entropy line and reweights the entropy values around it,
    approximating the decrease in entropy that would occur from observing that line.

    The neighbouring values are decreased by a Gaussian function centered at the selected line.
    """

    def __init__(
        self,
        n_actions: int,
        n_possible_actions: int,
        img_width: int,
        img_height: int,
        mean: float = 0,
        std_dev: float = 1,
        num_lines_to_update: int = 5,
    ):
        """Initialize the GreedyEntropy action selection model.

        Args:
            n_actions (int): The number of actions the agent can take.
            n_possible_actions (int): The number of possible actions.
            img_width (int): The width of the input image.
            img_height (int): The height of the input image.
            mean (float, optional): The mean of the RBF. Defaults to 0.
            std_dev (float, optional): The standard deviation of the RBF. Defaults to 1.
            num_lines_to_update (int, optional): The number of lines around the selected line
                to update. Must be odd.
        """
        super().__init__(n_actions, n_possible_actions, img_width, img_height)

        # Number of samples must be odd so that the entropy
        # of the selected line is set to 0 once it's been selected.
        assert num_lines_to_update % 2 == 1, "num_samples must be odd."
        self.num_lines_to_update = num_lines_to_update

        # see here what I mean by upside_down_gaussian:
        # https://colab.research.google.com/drive/1CQp_Z6nADzOFsybdiH5Cag0vtVZjjioU?usp=sharing
        upside_down_gaussian = lambda x: 1 - ops.exp(-0.5 * ((x - mean) / std_dev) ** 2)
        # Sample `num_lines_to_update` points symmetrically around the mean.
        # This can be tuned to determine how the entropy for neighbouring lines is updated
        # TODO: learn this function from training data
        points_to_evaluate = ops.linspace(
            mean - 2 * std_dev,
            mean + 2 * std_dev,
            self.num_lines_to_update,
        )
        self.upside_down_gaussian = upside_down_gaussian(points_to_evaluate)
        self.entropy_sigma = 1

    @staticmethod
    def compute_pairwise_pixel_gaussian_error(
        particles, stack_n_cols=1, n_possible_actions=None, entropy_sigma=1
    ):
        """Compute the pairwise pixelwise Gaussian error.

        This function computes the Gaussian error between each pair of pixels in the
        set of particles provided. This can be used to approximate the entropy of
        a Gaussian mixture model, where the particles are the means of the Gaussians.
        For more details see Section 4 here: https://arxiv.org/abs/2406.14388

        Args:
            particles (Tensor): Particles of shape (batch_size, n_particles, height, width)

        Returns:
            Tensor: batch of pixelwise pairwise Gaussian errors,
            of shape (n_particles, n_particles, batch, height, width)
        """
        assert particles.shape[1] > 1, "The entropy cannot be approximated using a single particle."

        if n_possible_actions is None:
            n_possible_actions = particles.shape[-1]

        # TODO: I think we only need to compute the lower triangular
        # of this matrix, since it's symmetric
        squared_l2_error_matrices = (particles[:, :, None, ...] - particles[:, None, :, ...]) ** 2
        gaussian_error_per_pixel_i_j = ops.exp((squared_l2_error_matrices) / (2 * entropy_sigma**2))
        # Vertically stack all columns corresponding with the same line
        # This way we can just sum across the height axis and get the entropy
        # for each pixel in a given line
        batch_size, n_particles, _, height, _ = gaussian_error_per_pixel_i_j.shape
        gaussian_error_per_pixel_stacked = ops.reshape(
            gaussian_error_per_pixel_i_j,
            [
                batch_size,
                n_particles,
                n_particles,
                height * stack_n_cols,
                n_possible_actions,
            ],
        )
        # [n_particles, n_particles, batch, height, width]
        return gaussian_error_per_pixel_stacked

    def compute_gmm_entropy_per_line(self, particles):
        """Compute the entropy for each line using a Gaussian Mixture Model.

        This function computes the entropy for each line using a Gaussian Mixture Model
        approximation of the posterior distribution.
        For more details see Section 4 here: https://arxiv.org/abs/2406.14388

        Args:
            particles (Tensor): Particles of shape (batch_size, n_particles, height, width)

        Returns:
            Tensor: batch of entropies per line, of shape (batch, n_possible_actions)
        """
        gaussian_error_per_pixel_stacked = GreedyEntropy.compute_pairwise_pixel_gaussian_error(
            particles,
            self.stack_n_cols,
            self.n_possible_actions,
            self.entropy_sigma,
        )
        gaussian_error_per_line = ops.sum(gaussian_error_per_pixel_stacked, axis=3)
        # sum out first dimension of (n_particles x n_particles) error matrix
        # [n_particles, batch, n_possible_actions]
        entropy_per_line_i = ops.sum(gaussian_error_per_line, axis=1)
        # sum out second dimension of (n_particles x n_particles) error matrix
        # [batch, n_possible_actions]
        entropy_per_line = ops.sum(entropy_per_line_i, axis=1)
        return entropy_per_line

    def select_line_and_reweight_entropy(self, entropy_per_line):
        """Select the line with maximum entropy and reweight the entropies.

        Selected the max entropy line and reweights the entropy values around it,
        approximating the decrease in entropy that would occur from observing that line.

        Args:
            entropy_per_line (Tensor): Entropy per line of shape
                (batch_size, n_possible_actions)

        Returns:
            Tuple: The selected line index and the updated entropies per line
        """

        # Find the line with maximum entropy
        max_entropy_line = ops.argmax(entropy_per_line)

        ## The rest of this function updates the entropy values around max_entropy_line
        ## by multiplying them with an upside-down Gaussian function centered at
        ## max_entropy_line, setting the entropy of the selected line to 0, and decreasing
        ## the entropies of neighbouring lines.

        # Pad the entropy per line to allow for re-weighting with fixed
        # size RBF, which is necessary for tracing.
        padded_entropy_per_line = ops.pad(
            entropy_per_line,
            (self.num_lines_to_update // 2, self.num_lines_to_update // 2),
        )
        # because the entropy per line has now been padded, the start index
        # of the set of lines to update is simply the index of the max_entropy_line
        start_index = max_entropy_line

        # Create the re-weighting vector
        reweighting = ops.ones_like(padded_entropy_per_line)
        reweighting = ops.slice_update(
            reweighting,
            (start_index,),
            ops.cast(self.upside_down_gaussian, dtype=reweighting.dtype),
        )

        # Apply re-weighting to entropy values
        updated_entropy_per_line_padded = padded_entropy_per_line * reweighting
        updated_entropy_per_line = ops.slice(
            updated_entropy_per_line_padded,
            (self.num_lines_to_update // 2,),
            (self.n_possible_actions,),
        )
        return max_entropy_line, updated_entropy_per_line

    def sample(self, particles):
        """Sample the action using the greedy entropy method.

        Args:
            particles (Tensor): Particles of shape (batch_size, n_particles, height, width)

        Returns:
           Tuple[Tensor, Tensor]:
                - Newly selected lines as k-hot vectors, shaped (batch_size, n_possible_actions)
                - Masks of shape (batch_size, img_height, img_width)
        """
        entropy_per_line = self.compute_gmm_entropy_per_line(particles)

        # Greedily select best line, reweight entropies, and repeat
        all_selected_lines = []
        for _ in range(self.n_actions):
            max_entropy_line, entropy_per_line = ops.vectorized_map(
                self.select_line_and_reweight_entropy, entropy_per_line
            )
            all_selected_lines.append(max_entropy_line)

        selected_lines_k_hot = ops.any(
            ops.one_hot(all_selected_lines, self.n_possible_actions, dtype=masks._DEFAULT_DTYPE),
            axis=0,
        )
        return selected_lines_k_hot, self.lines_to_im_size(selected_lines_k_hot)


@action_selection_registry(name="uniform_random")
class UniformRandomLines(LinesActionModel):
    """Uniform random lines action selection.

    Creates masks with uniformly randomly sampled lines.
    """

    def sample(self, batch_size=1, seed=None):
        """Sample the action using the uniform random method.

        Generates or updates an equispaced mask to sweep rightwards by one step across the image.

        Args:
            seed (int | SeedGenerator | jax.random.key, optional): Seed for random
                number generation. Defaults to None.

        Returns:
            Tuple[Tensor, Tensor]:
                - Newly selected lines as k-hot vectors, shaped (batch_size, n_possible_actions)
                - Masks of shape (batch_size, img_height, img_width)
        """
        selected_lines_batched = masks.random_uniform_lines(
            n_actions=self.n_actions,
            n_possible_actions=self.n_possible_actions,
            n_masks=batch_size,
            seed=seed,
        )
        return selected_lines_batched, self.lines_to_im_size(selected_lines_batched)


@action_selection_registry(name="equispaced")
class EquispacedLines(LinesActionModel):
    """Equispaced lines action selection.

    Creates masks with equispaced lines that sweep across
    the image.
    """

    def __init__(
        self,
        n_actions: int,
        n_possible_actions: int,
        img_width: int,
        img_height: int,
        assert_equal_spacing=True,
    ):
        super().__init__(n_actions, n_possible_actions, img_width, img_height)

        self.assert_equal_spacing = assert_equal_spacing
        if self.assert_equal_spacing:
            masks._assert_equal_spacing(n_actions, n_possible_actions)

    def sample(self, current_lines=None, batch_size=1):
        """Sample the action using the equispaced method.

        Generates or updates an equispaced mask to sweep rightwards by one step across the image.

        Returns:
            Tensor: The mask of shape (batch_size, img_size, img_size)
        """
        if current_lines is None:
            return self.initial_sample_stateless(batch_size)
        else:
            return self.sample_stateless(current_lines)

    def initial_sample_stateless(self, batch_size=1):
        """Initial sample stateless.

        Generates a batch of initial equispaced line masks.

        Returns:
            Tuple[Tensor, Tensor]:
                - Selected lines as k-hot vectors, shaped (batch_size, n_possible_actions)
                - Masks of shape (batch_size, img_height, img_width)
        """
        initial_lines = masks.initial_equispaced_lines(
            self.n_actions,
            self.n_possible_actions,
            assert_equal_spacing=self.assert_equal_spacing,
        )
        initial_lines = ops.tile(initial_lines, (batch_size, 1))  # (batch_size, n_actions)
        return initial_lines, self.lines_to_im_size(initial_lines)

    def sample_stateless(self, current_lines):
        """Sample stateless.

        Updates an existing equispaced mask to sweep rightwards by one step across the image.

        Args:
            current_lines: Currently selected lines as k-hot vectors,
                shaped (batch_size, n_possible_actions)

        Returns:
            Tuple[Tensor, Tensor]:
                - Newly selected lines as k-hot vectors, shaped (batch_size, n_possible_actions)
                - Masks of shape (batch_size, img_height, img_width)
        """
        new_lines = masks.next_equispaced_lines(current_lines)
        return new_lines, self.lines_to_im_size(new_lines)


@action_selection_registry(name="covariance")
class CovarianceSamplingLines(LinesActionModel):
    """Covariance sampling action selection.

    This class models the line-to-line correlation to select the mask with the highest entropy.
    """

    def __init__(
        self,
        n_actions: int,
        n_possible_actions: int,
        img_width: int,
        img_height: int,
        seed: int = 42,
        n_masks: int = 200,
    ):
        """Initialize the CovarianceSamplingLines action selection model.

        Args:
            n_actions (int): The number of actions the agent can take.
            n_possible_actions (int): The number of possible actions.
            img_width (int): The width of the input image.
            img_height (int): The height of the input image.
            seed (int, optional): The seed for random number generation. Defaults to 42.
            n_masks (int, optional): The number of masks. Defaults to 200.

        Raises:
            AssertionError: If image width is not divisible by n_possible_actions.
        """
        super().__init__(n_actions, n_possible_actions, img_width, img_height)
        self.seed = keras.random.SeedGenerator(seed)
        self.n_masks = n_masks

    def random_uniform_lines(self, batch_size, seed=None):
        """Wrapper around `random_uniform_lines` function to use attributes from class."""
        lines = masks.random_uniform_lines(
            self.n_actions,
            self.n_possible_actions,
            batch_size * self.n_masks,
            seed=self.seed if seed is None else seed,
        )
        return ops.reshape(lines, [self.n_masks, batch_size, self.n_possible_actions])

    def sample(self, particles, seed=None):
        """Sample the action using the covariance sampling method.

        Args:
            particles (Tensor): Particles of shape (batch_size, n_particles, h, w)
            seed (int | SeedGenerator | jax.random.key, optional): Seed for random number
                generation. Defaults to None.

        Returns:
            Tensor: The mask of shape (batch_size, img_size, img_size)
        """
        batch_size, n_particles, rows, _ = ops.shape(particles)

        # [batch_size, rows * stack_n_cols, n_possible_actions, n_particles]
        shape = [
            batch_size,
            rows * self.stack_n_cols,
            self.n_possible_actions,
            n_particles,
        ]
        particles = ops.reshape(particles, shape)

        # [batch_size, rows, n_possible_actions, n_possible_actions]
        cov_matrix = tensor_ops.batch_cov(particles)

        # Sum over the row dimension [batch_size, n_possible_actions, n_possible_actions]
        cov_matrix = ops.sum(cov_matrix, axis=1)

        # Generate random lines [n_masks, batch_size, n_possible_actions]
        lines = self.random_uniform_lines(batch_size, seed=seed)

        # Make matrix masks [n_masks, batch_size, n_possible_actions, n_possible_actions]
        reshaped_lines = ops.repeat(lines[..., None], self.n_possible_actions, axis=-1)
        bool_masks = ops.logical_and(reshaped_lines, ops.swapaxes(reshaped_lines, -1, -2))

        # Subsample the covariance matrix with random lines
        def subsample_with_mask(mask):
            """Subsample the covariance matrix with a single mask."""
            subsampled_cov_matrix = tensor_ops.boolean_mask(
                cov_matrix, mask, size=batch_size * self.n_actions**2
            )
            return ops.reshape(subsampled_cov_matrix, [batch_size, self.n_actions, self.n_actions])

        # [n_masks, batch_size, cols, cols]
        subsampled_cov_matrices = ops.vectorized_map(subsample_with_mask, bool_masks)

        # [n_masks, batch_size, 1]
        entropies = ops.logdet(subsampled_cov_matrices)[..., None]

        # [1, batch_size, 1]
        best_mask_index = ops.argmax(entropies, axis=0, keepdims=True)

        # [batch_size, n_possible_actions]
        best_mask = ops.take_along_axis(lines, best_mask_index, axis=0)
        best_mask = ops.squeeze(best_mask, axis=0)

        return best_mask, self.lines_to_im_size(best_mask)



"""
Identifies the scan cone of an ultrasound image and crops it such that the apex of the cone
is at the top center of the image. In this form, it can be scan converted to polar coordinates.

This module provides functionality to:
1. Detect the scan cone boundaries in ultrasound images
2. Fit lines to the cone edges
3. Calculate cone parameters (apex position, opening angle, etc.)
4. Crop and center the image around the cone
5. Visualize the detected cone and its parameters
"""

import argparse
import os
from pathlib import Path

if __name__ == "__main__":
    os.environ["KERAS_BACKEND"] = (
        "numpy"  # recommend using numpy for this since some line fitting is performed on CPU
    )

import keras
import matplotlib.pyplot as plt
import numpy as np
from keras import ops

from zea import log


def filter_edge_points_by_boundary(edge_points, is_left=True, min_cone_half_angle_deg=20):
    """
    Filter edge points to keep only those that form the actual boundary of the scan cone.
    Enforces minimum cone angle constraint to ensure valid cone shapes.

    Args:
        edge_points: Tensor of shape (N, 2) containing (x, y) coordinates of edge points
        is_left: Boolean indicating whether these are left (True) or right (False) edge points
        min_cone_half_angle_deg: Minimum expected half-angle of the cone in degrees

    Returns:
        Tensor of shape (M, 2) containing filtered edge points that satisfy the boundary constraints
    """
    if ops.shape(edge_points)[0] == 0:
        return edge_points

    # Convert minimum angle to slope
    min_slope = ops.tan(np.radians(min_cone_half_angle_deg))

    # Sort by y coordinate (top to bottom)
    sorted_indices = ops.argsort(edge_points[:, 1])
    sorted_points = ops.take(edge_points, sorted_indices, axis=0)

    filtered_points = []

    # Convert to numpy for the iterative logic (this part is hard to vectorize)
    sorted_points_np = ops.convert_to_numpy(sorted_points)

    for i, point in enumerate(sorted_points_np):
        x, y = point
        is_boundary_point = True

        # Check all points above this one
        for j in range(i):
            above_x, above_y = sorted_points_np[j]
            dy = y - above_y
            min_dx_required = min_slope * dy

            if is_left:
                # For left boundary
                if above_x < x or above_x - x < min_dx_required:
                    is_boundary_point = False
                    break
            else:
                # For right boundary
                if above_x > x or x - above_x < min_dx_required:
                    is_boundary_point = False
                    break

        if is_boundary_point:
            filtered_points.append(point)

    return ops.convert_to_tensor(filtered_points) if filtered_points else ops.zeros((0, 2))


def detect_cone_parameters(image, min_cone_half_angle_deg=20, threshold=15):
    """
    Detect the ultrasound cone parameters from a grayscale image.

    This function performs the following steps:
    1. Thresholds the image to create a binary mask
    2. Detects left and right edge points of the cone
    3. Filters edge points to ensure physical constraints
    4. Fits lines to the cone boundaries
    5. Calculates cone parameters including apex position, opening angle, and crop boundaries

    Args:
        image: 2D Keras tensor (grayscale image)
        min_cone_half_angle_deg: Minimum expected half-angle of the cone in degrees
        threshold: Threshold for binary image (pixels above this are considered data)

    Returns:
        Dictionary containing cone parameters including:
        - apex_x, apex_y: Coordinates of the cone apex
        - crop_left, crop_right, crop_top, crop_bottom: Crop boundaries
        - cone_height: Height of the cone
        - opening_angle: Opening angle of the cone
        - symmetry_ratio: Measure of cone symmetry
        - data_coverage: Fraction of crop region containing data
        - And other geometric parameters

    Raises:
        ValueError: If input image is not 2D or cone detection fails
    """
    try:
        import cv2
    except ImportError as exc:
        raise ImportError(
            "OpenCV is required for cone detection and visualization. "
            "Please install it with 'pip install opencv-python' or "
            "'pip install opencv-python-headless'."
        ) from exc

    if len(ops.shape(image)) != 2:
        raise ValueError("Input image must be 2D (grayscale)")

    # Ensure image is in proper range for cv2.threshold
    if image.dtype != "uint8":
        image = ops.cast(image * 255, "uint8")

    h, w = ops.shape(image)

    # OpenCV threshold requires numpy array
    image_np = ops.convert_to_numpy(image)
    _, thresh_np = cv2.threshold(image_np, threshold, 255, cv2.THRESH_BINARY)
    thresh = ops.convert_to_tensor(thresh_np)

    # Find non-zero pixel bounds
    non_zero_indices = ops.convert_to_tensor(ops.where(thresh > 0))
    if ops.shape(non_zero_indices)[0] == 0:
        return None

    min_y = ops.min(non_zero_indices[:, 0])
    max_y = ops.max(non_zero_indices[:, 0])

    # Collect left and right edge points
    left_edge_points = []
    right_edge_points = []

    # Convert back to numpy for row-wise processing (this is hard to vectorize efficiently)
    thresh_np = ops.convert_to_numpy(thresh)
    min_y_np = int(ops.convert_to_numpy(min_y))
    max_y_np = int(ops.convert_to_numpy(max_y))

    for y in range(min_y_np, max_y_np + 1):
        row = thresh_np[y, :]
        non_zero_x = np.where(row > 0)[0]

        if len(non_zero_x) > 0:
            width = non_zero_x[-1] - non_zero_x[0]
            if width > 5:  # Minimum width to avoid noise
                left_edge_points.append([non_zero_x[0], y])
                right_edge_points.append([non_zero_x[-1], y])

    if len(left_edge_points) < 10:
        return None

    left_edge_points = ops.convert_to_tensor(left_edge_points, dtype="float32")
    right_edge_points = ops.convert_to_tensor(right_edge_points, dtype="float32")

    # Filter edge points to keep only boundary points
    filtered_left_points = filter_edge_points_by_boundary(
        left_edge_points, is_left=True, min_cone_half_angle_deg=min_cone_half_angle_deg
    )
    filtered_right_points = filter_edge_points_by_boundary(
        right_edge_points,
        is_left=False,
        min_cone_half_angle_deg=min_cone_half_angle_deg,
    )

    if ops.shape(filtered_left_points)[0] < 3 or ops.shape(filtered_right_points)[0] < 3:
        return None

    # Fit lines using least squares: x = a + b*y
    # Convert to numpy for lstsq
    filtered_left_np = ops.convert_to_numpy(filtered_left_points)
    filtered_right_np = ops.convert_to_numpy(filtered_right_points)

    # Left line
    A_left = np.vstack([np.ones(len(filtered_left_np)), filtered_left_np[:, 1]]).T
    left_coeffs, _, _, _ = np.linalg.lstsq(A_left, filtered_left_np[:, 0], rcond=None)
    left_a, left_b = left_coeffs

    # Right line
    A_right = np.vstack([np.ones(len(filtered_right_np)), filtered_right_np[:, 1]]).T
    right_coeffs, _, _, _ = np.linalg.lstsq(A_right, filtered_right_np[:, 0], rcond=None)
    right_a, right_b = right_coeffs

    # Convert back to tensors
    left_a = ops.convert_to_tensor(left_a)
    left_b = ops.convert_to_tensor(left_b)
    right_a = ops.convert_to_tensor(right_a)
    right_b = ops.convert_to_tensor(right_b)

    # Calculate apex as intersection of fitted lines
    if ops.abs(left_b - right_b) < 1e-6:  # Lines are parallel
        return None

    apex_y = (right_a - left_a) / (left_b - right_b)
    apex_x = left_a + left_b * apex_y

    # Calculate cone height
    cone_height = max_y - apex_y

    # Calculate opening angle from the line slopes
    # Convert slopes to angles and calculate opening angle
    left_angle = ops.arctan(left_b)  # angle of left line from horizontal
    right_angle = ops.arctan(right_b)  # angle of right line from horizontal
    opening_angle = ops.abs(left_angle - right_angle)

    min_non_zero_pixel_idx = (0, 0)
    for i in reversed(range(0, thresh_np.shape[0])):
        row = thresh_np[i]
        non_zero_pixel_col = np.where(row > 0)[0]
        if np.any(non_zero_pixel_col):
            min_non_zero_pixel_idx = (i, non_zero_pixel_col[0])
            break

    circle_radius = float(
        np.sqrt(
            (min_non_zero_pixel_idx[1] - ops.convert_to_numpy(apex_x)) ** 2
            + (min_non_zero_pixel_idx[0] - ops.convert_to_numpy(apex_y)) ** 2
        )
    )
    circle_center_x = float(ops.convert_to_numpy(apex_x))
    circle_center_y = float(ops.convert_to_numpy(apex_y))

    # Calculate where the circle intersects with the cone lines
    # For line: x = a + b*y and circle: (x - cx)^2 + (y - cy)^2 = r^2
    def line_circle_intersection(a, b, cx, cy, r):
        """Find intersection of line x = a + b*y with circle centered at (cx, cy) with radius r"""
        # Substitute line equation into circle equation
        # (a + b*y - cx)^2 + (y - cy)^2 = r^2
        # Expand and collect terms to get quadratic in y
        A = 1 + b**2
        B = 2 * b * (a - cx) - 2 * cy
        C = (a - cx) ** 2 + cy**2 - r**2

        discriminant = B**2 - 4 * A * C
        if discriminant < 0:
            return None

        # Two solutions
        y1 = (-B + np.sqrt(discriminant)) / (2 * A)
        y2 = (-B - np.sqrt(discriminant)) / (2 * A)

        # Return the solution closest to the bottom (larger y)
        if y1 > y2:
            return y1, a + b * y1
        else:
            return y2, a + b * y2

    # Find intersection points
    left_intersect = line_circle_intersection(
        ops.convert_to_numpy(left_a),
        ops.convert_to_numpy(left_b),
        circle_center_x,
        circle_center_y,
        circle_radius,
    )
    right_intersect = line_circle_intersection(
        ops.convert_to_numpy(right_a),
        ops.convert_to_numpy(right_b),
        circle_center_x,
        circle_center_y,
        circle_radius,
    )

    if left_intersect is None or right_intersect is None:
        # Fallback to line endpoints at max_y
        left_y_bottom, left_x_bottom = (
            ops.convert_to_numpy(max_y),
            ops.convert_to_numpy(left_a + left_b * max_y),
        )
        right_y_bottom, right_x_bottom = (
            ops.convert_to_numpy(max_y),
            ops.convert_to_numpy(right_a + right_b * max_y),
        )
    else:
        left_y_bottom, left_x_bottom = left_intersect
        right_y_bottom, right_x_bottom = right_intersect

    sector_bottom = int(circle_radius + apex_y)  # this is the coordinate on the unpadded image

    # Calculate crop boundaries (can be negative)
    padding_x = 0
    padding_y = 0

    crop_left = int(left_x_bottom) - padding_x
    crop_right = int(right_x_bottom) + padding_x
    crop_top = int(ops.convert_to_numpy(apex_y)) - padding_y
    crop_bottom = int(sector_bottom)

    # Calculate final dimensions
    new_width = crop_right - crop_left
    new_height = crop_bottom - crop_top

    # Store the intersection points for visualization
    sector_left_x = left_x_bottom
    sector_left_y = left_y_bottom
    sector_right_x = right_x_bottom
    sector_right_y = right_y_bottom

    # Calculate symmetry ratio (how symmetric the cone is)
    symmetry_ratio = float(
        ops.convert_to_numpy(
            ops.abs(left_b + right_b) / (ops.abs(left_b) + ops.abs(right_b) + 1e-8)
        )
    )

    # Calculate data coverage in the crop region
    h_np = int(ops.convert_to_numpy(h))
    w_np = int(ops.convert_to_numpy(w))
    crop_left_clipped = max(0, crop_left)
    crop_right_clipped = min(w_np, crop_right)
    crop_top_clipped = max(0, crop_top)
    crop_bottom_clipped = min(h_np, crop_bottom)

    data_coverage = 0.0
    assert crop_left_clipped < crop_right_clipped and crop_top_clipped < crop_bottom_clipped
    crop_region = thresh_np[
        crop_top_clipped:crop_bottom_clipped, crop_left_clipped:crop_right_clipped
    ]
    data_coverage = float(np.sum(crop_region > 0) / crop_region.size)

    return {
        "apex_x": float(ops.convert_to_numpy(apex_x)),
        "apex_y": float(ops.convert_to_numpy(apex_y)),
        "crop_left": crop_left,
        "crop_right": crop_right,
        "crop_top": crop_top,
        "crop_bottom": crop_bottom,
        "original_width": int(ops.convert_to_numpy(w)),
        "original_height": int(ops.convert_to_numpy(h)),
        # Additional parameters for debug and analysis
        "cone_height": float(ops.convert_to_numpy(cone_height)),
        "opening_angle": float(ops.convert_to_numpy(opening_angle)),
        "new_width": new_width,
        "new_height": new_height,
        "symmetry_ratio": symmetry_ratio,
        "first_data_row": int(ops.convert_to_numpy(min_y)),
        "data_coverage": data_coverage,
        "apex_above_image": bool(ops.convert_to_numpy(apex_y) < 0),
        # Line parameters for reconstruction if needed
        "left_slope": float(ops.convert_to_numpy(left_b)),
        "right_slope": float(ops.convert_to_numpy(right_b)),
        "left_intercept": float(ops.convert_to_numpy(left_a)),
        "right_intercept": float(ops.convert_to_numpy(right_a)),
        # Circle parameters for the bottom boundary
        "circle_center_x": circle_center_x,
        "circle_center_y": circle_center_y,
        "circle_radius": circle_radius,
        # Sector intersection points for visualization
        "sector_left_x": sector_left_x,
        "sector_left_y": sector_left_y,
        "sector_right_x": sector_right_x,
        "sector_right_y": sector_right_y,
        "sector_bottom": sector_bottom,
    }


def crop_and_center_cone(image, cone_params):
    """
    Crop the image to the sector bounding box and pad as needed to center the apex.

    This function:
    1. Crops the image to the detected cone boundaries
    2. Adds padding if the apex is above the image
    3. Centers the apex horizontally in the final image

    Args:
        image: 2D Keras tensor (grayscale image)
        cone_params: Dictionary of cone parameters from detect_cone_parameters()

    Returns:
        Keras tensor of the cropped and centered image with the cone apex at the top center
    """
    # Get crop boundaries
    crop_left = cone_params["crop_left"]
    crop_right = cone_params["crop_right"]
    crop_top = cone_params["crop_top"]
    crop_bottom = cone_params["crop_bottom"]

    # Crop the image (handle negative crop_top)
    if crop_top < 0:
        cropped = image[0:crop_bottom, crop_left:crop_right]
        # Add top padding
        top_padding = -crop_top
        cropped_width = ops.shape(cropped)[1]
        top_pad = ops.zeros((top_padding, cropped_width), dtype=cropped.dtype)
        cropped = ops.concatenate([top_pad, cropped], axis=0)
    else:
        cropped = image[crop_top:crop_bottom, crop_left:crop_right]

    # Now handle horizontal centering
    # Calculate where the apex is in the cropped image
    apex_x_in_crop = cone_params["apex_x"] - crop_left
    cropped_height = ops.shape(cropped)[0]
    cropped_width = ops.shape(cropped)[1]

    # Calculate the target center position
    target_center_x = ops.cast(cropped_width / 2, "float32")

    # Calculate how much padding we need on each side
    # We want: left_padding + apex_x_in_crop = final_width / 2
    # And: final_width = cropped_width + left_padding + right_padding
    # For symmetric padding: left_padding = right_padding
    # So: left_padding + apex_x_in_crop = (cropped_width + 2*left_padding) / 2
    # Solving: left_padding = cropped_width/2 - apex_x_in_crop

    left_padding_needed = target_center_x - apex_x_in_crop

    # Ensure we have non-negative padding
    left_padding = ops.maximum(0, ops.cast(left_padding_needed, "int32"))
    right_padding = ops.maximum(0, ops.cast(-left_padding_needed, "int32"))

    # Apply horizontal padding if needed
    if left_padding > 0 or right_padding > 0:
        if left_padding > 0:
            left_pad = ops.zeros((cropped_height, left_padding), dtype=cropped.dtype)
            cropped = ops.concatenate([left_pad, cropped], axis=1)

        if right_padding > 0:
            right_pad = ops.zeros((cropped_height, right_padding), dtype=cropped.dtype)
            cropped = ops.concatenate([cropped, right_pad], axis=1)

    return cropped


def fit_and_crop_around_scan_cone(
    image_tensor, min_cone_half_angle_deg=20, threshold=15, return_params=False
):
    """
    Detect scan cone in ultrasound image and return cropped/padded image with centered apex.

    Args:
        image_tensor: Keras tensor (2D grayscale image)
        min_cone_half_angle_deg: Minimum expected half-angle of the cone in degrees (default: 20)
        threshold: Threshold for binary image - pixels above this are considered data (default: 15)
        return_params: If True, also return cone parameters (default: False)

    Returns:
        - If return_params is False: Keras tensor (cropped and padded image with apex at center)
        - If return_params is True: Tuple of (cropped_tensor, cone_parameters_dict)

    Raises:
        ValueError: If cone detection fails or image is not 2D
    """
    if keras.backend.backend() != "numpy":
        log.info(f"❗️ It is recommended to use {log.blue('numpy')} backend for `fit_scan_cone()`.")

    # Ensure image is 2D
    if len(ops.shape(image_tensor)) != 2:
        raise ValueError(f"Input must be 2D grayscale image, got shape {ops.shape(image_tensor)}")

    # Detect cone parameters
    cone_params = detect_cone_parameters(
        image_tensor,
        min_cone_half_angle_deg=min_cone_half_angle_deg,
        threshold=threshold,
    )

    if cone_params is None:
        raise ValueError("Failed to detect ultrasound cone in image")

    # Crop and center the image
    cropped_image = crop_and_center_cone(image_tensor, cone_params)

    if return_params:
        return cropped_image, cone_params
    else:
        return cropped_image


def visualize_scan_cone(image, cone_params, output_dir="output"):
    """
    Create visualization plots for the scan cone detection.

    Args:
        image: Original grayscale image
        cone_params: Dictionary of cone parameters from detect_cone_parameters()
        output_dir: Directory to save output plots (default: "output")

    The visualization is saved as 'scan_cone_visualization.png' in the output directory.
    """
    # Create output directory
    output_path = Path(output_dir)
    output_path.mkdir(parents=True, exist_ok=True)

    # Create figure
    _, ax = plt.subplots(1, 1, figsize=(12, 10))
    ax.imshow(image, cmap="gray" if len(image.shape) == 2 else None)

    # Extract parameters
    apex_x, apex_y = cone_params["apex_x"], cone_params["apex_y"]
    left_a = cone_params["left_intercept"]
    left_b = cone_params["left_slope"]
    right_a = cone_params["right_intercept"]
    right_b = cone_params["right_slope"]

    # Get bottom intersection points
    bottom_y = image.shape[0] - 1
    left_x_bottom = left_a + left_b * bottom_y
    right_x_bottom = right_a + right_b * bottom_y

    # Draw cone lines
    ax.plot(
        [apex_x, left_x_bottom],
        [apex_y, bottom_y],
        color="#FF0000",  # Bright red
        linewidth=2,
        label="Cone boundary",
    )
    ax.plot(
        [apex_x, right_x_bottom],
        [apex_y, bottom_y],
        color="#FF0000",  # Bright red
        linewidth=2,
    )

    # Draw circular arc between the fitted lines
    if "circle_center_x" in cone_params:
        circle_center_x = cone_params["circle_center_x"]
        circle_center_y = cone_params["circle_center_y"]
        circle_radius = cone_params["circle_radius"]

        # Calculate angles for left and right intersection points
        left_angle = np.arctan2(left_x_bottom - circle_center_x, bottom_y - circle_center_y)
        right_angle = np.arctan2(right_x_bottom - circle_center_x, bottom_y - circle_center_y)

        # Ensure angles are in the correct order
        if right_angle < left_angle:
            right_angle += 2 * np.pi

        # Create arc points
        theta = np.linspace(left_angle, right_angle, 100)
        arc_x = circle_center_x + circle_radius * np.sin(theta)
        arc_y = circle_center_y + circle_radius * np.cos(theta)

        # Only draw the part that's visible in the image
        visible_mask = (
            (arc_y >= 0) & (arc_y < image.shape[0]) & (arc_x >= 0) & (arc_x < image.shape[1])
        )
        if np.any(visible_mask):
            ax.plot(
                arc_x[visible_mask],
                arc_y[visible_mask],
                color="#00FF00",  # Bright green
                linestyle="--",
                linewidth=2,
                label="Sector arc",
            )

    # Draw apex as a star
    ax.plot(
        apex_x,
        apex_y,
        marker="*",  # Star marker
        markersize=15,
        color="#FFD700",  # Gold
        # markeredgecolor="white",
        markeredgewidth=2,
        label="Cone apex",
    )

    # Draw crop rectangle
    rect = plt.Rectangle(
        (cone_params["crop_left"], cone_params["crop_top"]),
        cone_params["crop_right"] - cone_params["crop_left"],
        cone_params["crop_bottom"] - cone_params["crop_top"],
        fill=False,
        color="#00FFFF",  # Cyan
        linewidth=2,
        label="Crop region",
        linestyle="dotted",
    )
    ax.add_patch(rect)

    # Add info text
    opening_angle_deg = np.degrees(cone_params["opening_angle"])
    info_text = (
        f"Opening Angle: {opening_angle_deg:.1f}°\n"
        f"Symmetry Ratio: {cone_params.get('symmetry_ratio', 0.0):.3f}"
    )

    if "circle_radius" in cone_params:
        info_text += f"\nCircle Radius: {cone_params['circle_radius']:.1f} px"

    ax.text(
        0.02,
        0.98,
        info_text,
        transform=ax.transAxes,
        fontsize=11,
        va="top",
        bbox={
            "boxstyle": "round",
            "facecolor": "#FFFF00",
            "alpha": 0.9,
        },  # Bright yellow
    )

    # Add crop info
    crop_info = (
        f"Crop: ({cone_params['crop_left']}, {cone_params['crop_top']}) to "
        f"({cone_params['crop_right']}, {cone_params['crop_bottom']})\n"
        f"Final Size: {cone_params['new_width']} x {cone_params['new_height']}"
    )

    ax.text(
        0.02,
        0.02,
        crop_info,
        transform=ax.transAxes,
        fontsize=11,
        va="bottom",
        bbox={
            "boxstyle": "round",
            "facecolor": "#00FFFF",
            "alpha": 0.9,
        },
    )

    # Set up the plot
    ax.set_xlim(0, image.shape[1])
    ax.set_ylim(image.shape[0], 0)
    ax.legend(loc="upper right")
    ax.set_title("Scan Cone Detection", fontsize=16)
    ax.axis("off")

    plt.tight_layout()
    plt.savefig(
        output_path / "scan_cone_visualization.png",
        dpi=300,
        bbox_inches="tight",
        pad_inches=0.5,
    )
    log.info(
        f"Saved cone visualization to {log.yellow(output_path / 'scan_cone_visualization.png')}"
    )
    plt.close()


def main(avi_path):
    """Demonstrate scan cone fitting on a sample AVI file."""
    try:
        import cv2
    except ImportError as exc:
        raise ImportError(
            "OpenCV is required for cone detection and visualization. "
            "Please install it with 'pip install opencv-python' or "
            "'pip install opencv-python-headless'."
        ) from exc

    # Load first frame
    cap = cv2.VideoCapture(avi_path)
    ret, frame = cap.read()
    cap.release()

    if not ret:
        print(f"Failed to read video file: {avi_path}")
        return

    # Convert to grayscale
    frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)

    # Convert to tensor
    frame_tensor = ops.convert_to_tensor(frame)

    try:
        # Fit scan cone
        _, cone_params = fit_and_crop_around_scan_cone(
            frame_tensor, min_cone_half_angle_deg=20, threshold=15, return_params=True
        )

        # Create visualization
        visualize_scan_cone(frame, cone_params)
        print("Visualization saved to output/scan_cone_visualization.png")

    except ValueError as e:
        print(f"Error fitting scan cone: {e}")


def get_args():
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(
        description="Precompute cone parameters for EchoNet-LVH dataset"
    )
    parser.add_argument(
        "--input_file",
        type=str,
        required=True,
    )
    return parser.parse_args()


if __name__ == "__main__":
    args = get_args()
    input_file = args.input_file
    main(input_file)



"""
H5 dataloader for loading images from zea datasets.
"""

import re
from itertools import product
from pathlib import Path
from typing import List

import numpy as np

from zea import log
from zea.data.datasets import Dataset, H5FileHandleCache, count_samples_per_directory
from zea.data.file import File
from zea.data.utils import json_dumps
from zea.io_lib import retry_on_io_error
from zea.utils import map_negative_indices

DEFAULT_NORMALIZATION_RANGE = (0, 1)
MAX_RETRY_ATTEMPTS = 3
INITIAL_RETRY_DELAY = 0.1


def generate_h5_indices(
    file_paths: List[str],
    file_shapes: list,
    n_frames: int,
    frame_index_stride: int,
    key: str = "data/image",
    initial_frame_axis: int = 0,
    additional_axes_iter: List[int] | None = None,
    sort_files: bool = True,
    overlapping_blocks: bool = False,
    limit_n_frames: int | None = None,
):
    """Generate indices for h5 files.

    Generates a list of indices to extract images from hdf5 files. Length of this list
    is the length of the extracted dataset.

    Args:
        file_paths (list): List of file paths.
        file_shapes (list): List of file shapes.
        n_frames (int): Number of frames to load from each hdf5 file.
        frame_index_stride (int): Interval between frames to load.
        key (str, optional): Key of hdf5 dataset to grab data from. Defaults to "data/image".
        initial_frame_axis (int, optional): Axis to iterate over. Defaults to 0.
        additional_axes_iter (list, optional): Additional axes to iterate over in the dataset.
            Defaults to None.
        sort_files (bool, optional): Sort files by number. Defaults to True.
        overlapping_blocks (bool, optional): Will take n_frames from sequence, then move by 1.
            Defaults to False.
        limit_n_frames (int, optional): Limit the number of frames to load from each file. This
            means n_frames per data file will be used. These will be the first frames in the file.
            Defaults to None.

    Returns:
        list: List of tuples with indices to extract images from hdf5 files.
            (file_name, key, indices) with indices being a tuple of slices.

    Example:
        .. code-block:: python

            [
                (
                    "/folder/path_to_file.hdf5",
                    "data/image",
                    [range(0, 1), slice(None, 256, None), slice(None, 256, None)],
                ),
                (
                    "/folder/path_to_file.hdf5",
                    "data/image",
                    [range(1, 2), slice(None, 256, None), slice(None, 256, None)],
                ),
                ...,
            ]
    """
    if not limit_n_frames:
        limit_n_frames = np.inf

    assert len(file_paths) == len(file_shapes), "file_paths and file_shapes must have same length"

    if additional_axes_iter:
        # cannot contain initial_frame_axis
        assert initial_frame_axis not in additional_axes_iter, (
            "initial_frame_axis cannot be in additional_axes_iter. "
            "We are already iterating over that axis."
        )
    else:
        additional_axes_iter = []

    if sort_files:
        try:
            # this is like an np.argsort, returns the indices that would sort the array
            indices_sorting_file_paths = sorted(
                range(len(file_paths)),
                key=lambda i: int(re.findall(r"\d+", file_paths[i])[-2]),
            )
            file_paths = [file_paths[i] for i in indices_sorting_file_paths]
            file_shapes = [file_shapes[i] for i in indices_sorting_file_paths]
        except Exception:
            log.warning("H5Generator: Could not sort file_paths by number.")

    # block size with stride included
    block_size = n_frames * frame_index_stride

    if not overlapping_blocks:
        block_step_size = block_size
    else:
        # now blocks overlap by n_frames - 1
        block_step_size = 1

    def axis_indices_files():
        # For every file
        for shape in file_shapes:
            n_frames_in_file = shape[initial_frame_axis]
            # Optionally limit frames to load from each file
            n_frames_in_file = min(n_frames_in_file, limit_n_frames)
            indices = [
                range(i, i + block_size, frame_index_stride)
                for i in range(0, n_frames_in_file - block_size + 1, block_step_size)
            ]
            yield [indices]

    indices = []
    skipped_files = 0
    for file, shape, axis_indices in zip(file_paths, file_shapes, list(axis_indices_files())):
        # remove all the files that have empty list at initial_frame_axis
        # this can happen if the file is too small to fit a block
        if not axis_indices[0]:  # initial_frame_axis is the first entry in axis_indices
            skipped_files += 1
            continue

        if additional_axes_iter:
            axis_indices += [range(shape[axis]) for axis in additional_axes_iter]

        axis_indices = product(*axis_indices)

        for axis_index in axis_indices:
            full_indices = [slice(size) for size in shape]
            for i, axis in enumerate([initial_frame_axis] + list(additional_axes_iter)):
                full_indices[axis] = axis_index[i]
            indices.append((file, key, full_indices))

    if skipped_files > 0:
        log.warning(
            f"H5Generator: Skipping {skipped_files} files with not enough frames "
            f"which is about {skipped_files / len(file_paths) * 100:.2f}% of the "
            f"dataset. This can be fine if you expect set `n_frames` and "
            "`frame_index_stride` to be high. Minimum frames in a file needs to be at "
            f"least n_frames * frame_index_stride = {n_frames * frame_index_stride}. "
        )

    return indices


def _h5_reopen_on_io_error(
    dataloader_obj: H5FileHandleCache,
    file,
    key,
    indices,
    retry_count,
    **kwargs,
):
    """Reopen the file if an I/O error occurs.
    Also removes the file from the cache and try to close file.
    """
    file_name = indices[0]
    try:
        file_handle = dataloader_obj._file_handle_cache.pop(file_name, None)
        if file_handle is not None:
            file_handle.close()
    except Exception:
        pass

    log.warning(
        f"H5Generator: I/O error occurred while reading file {file_name}. "
        f"Retry opening file. Retry count: {retry_count}."
    )


class H5Generator(Dataset):
    """H5Generator class for iterating over hdf5 files in an advanced way.
    Mostly used internally, you might want to use the Dataloader class instead.
    Loads one item at a time. Always outputs numpy arrays.
    """

    def __init__(
        self,
        file_paths: List[str],
        key: str = "data/image",
        n_frames: int = 1,
        shuffle: bool = True,
        return_filename: bool = False,
        limit_n_samples: int | None = None,
        limit_n_frames: int | None = None,
        seed: int | None = None,
        cache: bool = False,
        additional_axes_iter: tuple | None = None,
        sort_files: bool = True,
        overlapping_blocks: bool = False,
        initial_frame_axis: int = 0,
        insert_frame_axis: bool = True,
        frame_index_stride: int = 1,
        frame_axis: int = -1,
        validate: bool = True,
        **kwargs,
    ):
        super().__init__(file_paths, key, validate=validate, **kwargs)

        self.n_frames = int(n_frames)
        self.frame_index_stride = int(frame_index_stride)
        self.frame_axis = int(frame_axis)
        self.insert_frame_axis = insert_frame_axis
        self.initial_frame_axis = int(initial_frame_axis)
        self.return_filename = return_filename
        self.shuffle = shuffle
        self.sort_files = sort_files
        self.overlapping_blocks = overlapping_blocks
        self.limit_n_samples = limit_n_samples
        self.limit_n_frames = limit_n_frames
        self.seed = seed
        self.additional_axes_iter = additional_axes_iter or []

        assert self.frame_index_stride > 0, (
            f"`frame_index_stride` must be greater than 0, got {self.frame_index_stride}"
        )
        assert self.n_frames > 0, f"`n_frames` must be greater than 0, got {self.n_frames}"

        # Extract some general information about the dataset
        image_shapes = np.array(self.file_shapes)
        image_shapes = np.delete(
            image_shapes, (self.initial_frame_axis, *self.additional_axes_iter), axis=1
        )
        n_dims = len(image_shapes[0])

        self.equal_file_shapes = np.all(image_shapes == image_shapes[0])
        if not self.equal_file_shapes:
            log.warning(
                "H5Generator: Not all files have the same shape. "
                "This can lead to issues when resizing images later...."
            )
            self.shape = np.array([None] * n_dims)
        else:
            self.shape = np.array(image_shapes[0])

        if insert_frame_axis:
            _frame_axis = map_negative_indices([frame_axis], len(self.shape) + 1)
            self.shape = np.insert(self.shape, _frame_axis, 1)
        if self.shape[frame_axis]:
            self.shape[frame_axis] = self.shape[frame_axis] * n_frames

        # Set random number generator
        self.rng = np.random.default_rng(self.seed)

        self.indices = generate_h5_indices(
            file_paths=self.file_paths,
            file_shapes=self.file_shapes,
            n_frames=self.n_frames,
            frame_index_stride=self.frame_index_stride,
            key=self.key,
            initial_frame_axis=self.initial_frame_axis,
            additional_axes_iter=self.additional_axes_iter,
            sort_files=self.sort_files,
            overlapping_blocks=self.overlapping_blocks,
            limit_n_frames=self.limit_n_frames,
        )

        if not self.shuffle:
            log.warning("H5Generator: Not shuffling data.")

        if limit_n_samples:
            log.warning(
                f"H5Generator: Limiting number of samples to {limit_n_samples} "
                f"out of {len(self.indices)}"
            )
            self.indices = self.indices[:limit_n_samples]

        self.shuffled_items = list(range(len(self.indices)))

        # Retry count for I/O errors
        self.retry_count = 0

        # Create a cache for the data
        self.cache = cache
        self._data_cache = {}

    def _get_single_item(self, idx):
        # Check if the item is already in the cache
        if self.cache and idx in self._data_cache:
            return self._data_cache[idx]

        # Get the data
        file_name, key, indices = self.indices[idx]
        file = self.get_file(file_name)
        image = self.load(file, key, indices)
        file_data = json_dumps(
            {
                "fullpath": file.filename,
                "filename": file.stem,
                "indices": indices,
            }
        )

        if self.cache:
            # Store the image and file data in the cache
            self._data_cache[idx] = [image, file_data]

        return image, file_data

    def __getitem__(self, index):
        image, file_data = self._get_single_item(self.shuffled_items[index])

        if self.return_filename:
            return image, file_data
        else:
            return image

    @retry_on_io_error(
        max_retries=MAX_RETRY_ATTEMPTS,
        initial_delay=INITIAL_RETRY_DELAY,
        retry_action=_h5_reopen_on_io_error,
    )
    def load(self, file: File, key: str, indices: tuple | str):
        """Extract data from hdf5 file.
        Args:
            file_name (str): name of the file to extract image from.
            key (str): key of the hdf5 dataset to grab data from.
            indices (tuple): indices to extract image from (tuple of slices)
        Returns:
            np.ndarray: image extracted from hdf5 file and indexed by indices.
        """
        try:
            images = file.load_data(key, indices)
        except (OSError, IOError):
            # Let the decorator handle I/O errors
            raise
        except Exception as exc:
            # For non-I/O errors, provide detailed context
            raise ValueError(
                f"Could not load image at index {indices} "
                f"and file {file.name} of shape {file[key].shape}"
            ) from exc

        # stack frames along frame_axis
        if self.insert_frame_axis:
            # move frames axis to self.frame_axis
            initial_frame_axis = self.initial_frame_axis
            if self.additional_axes_iter:
                # offset initial_frame_axis if we have additional axes that are before
                # the initial_frame_axis
                additional_axes_before = sum(
                    axis < self.initial_frame_axis for axis in self.additional_axes_iter
                )
                initial_frame_axis = initial_frame_axis - additional_axes_before

            images = np.moveaxis(images, initial_frame_axis, self.frame_axis)
        else:
            # append frames to existing axis
            images = np.concatenate(images, axis=self.frame_axis)

        return images

    def _shuffle(self):
        self.rng.shuffle(self.shuffled_items)
        log.info("H5Generator: Shuffled data.")

    def __len__(self):
        return len(self.indices)

    def iterator(self):
        """Generator that yields images from the hdf5 files."""
        if self.shuffle:
            self._shuffle()
        for idx in range(len(self)):
            yield self[idx]

    def __iter__(self):
        """
        Generator that yields images from the hdf5 files.
        """
        return self.iterator()

    def __repr__(self):
        return (
            f"<{self.__class__.__name__} at 0x{id(self):x}: "
            f"{len(self)} batches, n_frames={self.n_frames}, key='{self.key}', "
            f"shuffle={self.shuffle}, file_paths={len(self.file_paths)}>"
        )

    def __str__(self):
        return (
            f"H5Generator with {len(self)} batches from {len(self.file_paths)} files "
            f"(key='{self.key}')"
        )

    def summary(self):
        """Return a string with dataset statistics and per-directory breakdown."""
        total_samples = len(self.indices)
        file_names = [idx[0] for idx in self.indices]
        # Try to infer directories from file_names
        directories = sorted({str(Path(f).parent) for f in file_names})
        samples_per_dir = count_samples_per_directory(file_names, directories)

        parts = [f"H5Generator with {total_samples} total samples:"]
        for dir_path, count in samples_per_dir.items():
            percentage = (count / total_samples) * 100 if total_samples else 0
            parts.append(f"  {dir_path}: {count} samples ({percentage:.1f}%)")
        print("\n".join(parts))



"""Parameter descriptions for the config file."""

from zea.internal.config.validation import _ALLOWED_PLOT_LIBS, _DATA_TYPES


def allows_type_to_str(allowed_types):
    """Transforms a list of allowed types into a string for use in a comment."""
    ouput_str = ", ".join([str(a) if a is not None else "null" for a in allowed_types])
    return ouput_str


PARAMETER_DESCRIPTIONS = {
    "data": {
        "description": "The data section contains the parameters for the data.",
        "dataset_folder": (
            "The path of the folder to load data files from (relative to the user data "
            "root as set in users.yaml)"
        ),
        "to_dtype": (f"The type of data to convert to ({allows_type_to_str(_DATA_TYPES)})"),
        "file_path": (
            "The path of the file to load when running the UI (either an absolute path "
            "or one relative to the dataset folder)"
        ),
        "frame_no": "The frame number to load when running the UI (null, int, 'all')",
        "input_range": "The range of the input data in db (null, [min, max])",
        "apodization": "The receive apodization to use.",
        "output_range": ("The output range to which the data should be mapped (e.g. [0, 1])."),
        "resolution": ("The spatial resolution of the data in meters per pixel (float, optional)."),
        "local": "true: use local data on this device, false: use data from NAS",
        "dtype": (
            "The form of data to load (raw_data, rf_data, iq_data, beamformed_data, "
            "envelope_data, image, image_sc)"
        ),
        "dynamic_range": "The dynamic range for showing data in db [min, max]",
        "user": "The user to use when loading data (null, dict)",
    },
    "scan": {
        "description": (
            "The scan section contains the parameters pertaining to the reconstruction."
        ),
        "selected_transmits": (
            "The number of transmits in a frame. Can be 'all' for all transmits, an "
            "integer for a specific number of transmits selected evenly from the "
            "transmits in the frame, or a list of integers for specific transmits to "
            "select from the frame."
        ),
        "Nx": "The number of pixels in the beamforming grid in the x-direction",
        "Nz": "The number of pixels in the beamforming grid in the z-direction",
        "n_ch": "The number of channels in the raw data (1 for rf data, 2 for iq data)",
        "n_ax": "The number of samples in a receive recording per channel.",
        "xlims": "The limits of the x-axis in the scan in meters (null, [min, max])",
        "ylims": "The limits of the y-axis in the scan in meters (null, [min, max])",
        "zlims": "The limits of the z-axis in the scan in meters (null, [min, max])",
        "center_frequency": "The center frequency of the transducer in Hz",
        "sampling_frequency": "The sampling frequency of the data in Hz",
        "demodulation_frequency": (
            "The demodulation frequency of the data in Hz. This is the assumed center "
            "frequency of the transmit waveform used to demodulate the rf data to iq "
            "data."
        ),
        "apply_lens_correction": (
            "Set to true to apply lens correction in the time-of-flight calculation"
        ),
        "lens_thickness": "The thickness of the lens in meters",
        "lens_sound_speed": ("The speed of sound in the lens in m/s. Usually around 1000 m/s"),
        "f_number": (
            "The receive f-number for apodization. Set to zero to disable masking. "
            "The f-number is the ratio between the distance from the transducer and the "
            "size of the aperture."
        ),
        "fill_value": (
            "Value to fill the image with outside the defined region (float, default 0.0)."
        ),
        "phi_range": (
            "The range of phi values in radians for 3D scan conversion (null, [min, max])."
        ),
        "theta_range": (
            "The range of theta values in radians for scan conversion (null, [min, max])."
        ),
        "rho_range": ("The range of rho values in meters for scan conversion (null, [min, max])."),
        "resolution": ("The resolution for scan conversion in meters per pixel (float, optional)."),
    },
    "pipeline": {
        "description": "This section contains the necessary parameters for building the pipeline.",
        "operations": (
            "The operations to perform on the data. This is a list of dictionaries, "
            "where each dictionary contains the parameters for a single operation."
        ),
        "with_batch_dim": (
            "Whether operations should expect a batch dimension in the input. Defaults to True."
        ),
        "jit_options": (
            "The JIT options to use. Must be 'pipeline', 'ops', or None. "
            "'pipeline' compiles the entire pipeline as a single function. "
            "'ops' compiles each operation separately. None disables JIT compilation. "
            "Defaults to 'ops'."
        ),
        "jit_kwargs": ("Additional keyword arguments for the JIT compiler. Defaults to None."),
        "name": ("The name of the pipeline. Defaults to 'pipeline'."),
        "validate": ("Whether to validate the pipeline. Defaults to True."),
    },
    "device": "The device to run on ('cpu', 'gpu:0', 'gpu:1', ...)",
    "plot": {
        "description": (
            "Settings pertaining to plotting when running the UI "
            "(`zea --config <path-to-config.yaml>`)"
        ),
        "save": ("Set to true to save the plots to disk, false to only display them in the UI"),
        "plot_lib": (f"The plotting library to use ({allows_type_to_str(_ALLOWED_PLOT_LIBS)})"),
        "fps": "Frames per second for video output.",
        "tag": "The name for the plot",
        "fliplr": "Set to true to flip the image left to right",
        "image_extension": "The file extension to use when saving the image (png, jpg)",
        "video_extension": "The file extension to use when saving the video (mp4, gif)",
        "headless": "Set to true to run the UI in headless mode",
        "selector": (
            "Type of selector to use for ROI selection in the UI ('rectangle', 'lasso', or None)."
        ),
        "selector_metric": ("Metric to use for evaluating selected regions (e.g., 'gcnr')."),
    },
    "git": "The git commit hash or branch for reproducibility (string, optional).",
    "hide_devices": ("List of device indices to hide from selection (list of int, optional)."),
}



"""Base classes for the toolbox"""

import enum
import json
import pickle
from copy import deepcopy

import keras
import numpy as np

from zea.utils import reduce_to_signature, update_dictionary

CONVERT_TO_KERAS_TYPES = (np.ndarray, int, float, list, tuple, bool)
BASE_FLOAT_PRECISION = "float32"
BASE_INT_PRECISION = "int32"

# TODO: make static more neat
# These are global static attributes for all ops. Ops specific
# static attributes should be defined in the respective ops class
STATIC = [
    "f_number",
    "apply_lens_correction",
    "apply_phase_rotation",
    "Nx",
    "Nz",
    "fill_value",
]


class DataTypes(enum.Enum):
    """Enum class for zea data types.

    The following terminology is used in the code when referring to different
    data types.

    raw_data        --> The raw channel data, storing the time-samples from each
                        distinct ultrasound transducer.
    aligned_data    --> Time-of-flight (TOF) corrected data. This is the data
                        that is time aligned with respect to the array geometry.
    beamformed_data --> Beamformed or also known as beamsummed data. Aligned
                        data is coherently summed together along the elements.
                        The data has now been transformed from the aperture
                        domain to the spatial domain.
    envelope_data   --> The envelope of the signal is here detected and the
                        center frequency is removed from the signal.
    image           --> After log compression of the envelope data, the
                        image is formed.
    image_sc        --> The scan converted image is transformed to cartesian
                        (x, y) format to account for possible curved arrays.

    """

    RAW_DATA = "raw_data"
    ALIGNED_DATA = "aligned_data"
    BEAMFORMED_DATA = "beamformed_data"
    ENVELOPE_DATA = "envelope_data"
    IMAGE = "image"
    IMAGE_SC = "image_sc"


class ModTypes(enum.Enum):
    """Enum class for zea modulation types."""

    RF = "rf"
    IQ = "iq"
    NONE = None


class classproperty(property):
    """Define a class level property."""

    def __get__(self, _, owner_cls):
        return self.fget(owner_cls)


class Object:
    """Base class for all data objects in the toolbox"""

    def __init__(self):
        self._serialized = None

    @property
    def serialized(self):
        """Compute the checksum of the object only if not already done"""
        if self._serialized is None:
            attributes = self.__dict__.copy()
            attributes.pop(
                "_serialized", None
            )  # Remove the cached serialized attribute to avoid recursion
            self._serialized = pickle.dumps(attributes)
        return self._serialized

    def __setattr__(self, name: str, value):
        """Reset the serialized data if the object is modified"""
        if name != "_serialized":  # Avoid resetting when setting _serialized itself
            self._serialized = None
        super().__setattr__(name, value)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self.serialized == other.serialized

    def __hash__(self):
        return hash(self.serialized)

    def copy(self):
        """Return a copied version of the object"""
        return deepcopy(self)

    def update(self, **kwargs):
        """Update the attributes of the object if they exist"""
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)

    def __getitem__(self, key):
        return getattr(self, key)

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def __delitem__(self, key):
        delattr(self, key)

    def to_tensor(self):
        """Convert the attributes in the object to keras tensors"""
        return object_to_tensor(self)

    @classmethod
    def safe_initialize(cls, **kwargs):
        """Safely initialize a class by removing any invalid arguments."""
        reduced_params = reduce_to_signature(cls.__init__, kwargs)
        return cls(**reduced_params)

    @classmethod
    def merge(cls, obj1: dict, obj2: dict):
        """Merge multiple objects and safely initialize a new object."""
        # TODO: support actual zea.core.Objects, now we only support dictionaries
        params = update_dictionary(obj1, obj2)
        return cls.safe_initialize(**params)

    @classmethod
    def _tree_unflatten(cls, aux, children):
        if cls is not Object:
            raise NotImplementedError(f"{cls.__name__} must implement _tree_unflatten.")
        return cls(*children)

    def _tree_flatten(self):
        if not isinstance(self, Object):
            raise NotImplementedError(f"{type(self).__name__} must implement _tree_flatten.")
        return (), ()

    @classmethod
    def register_pytree_node(cls):
        """Register the object as a PyTree node for JAX.
        https://docs.jax.dev/en/latest/_autosummary/jax.tree_util.register_pytree_node.html
        """
        try:
            from jax import tree_util
        except ImportError as exc:
            raise ImportError(
                "JAX is not installed. Please install JAX to use `register_pytree_node`."
            ) from exc

        tree_util.register_pytree_node(
            cls,
            cls._tree_flatten,
            cls._tree_unflatten,
        )


def object_to_tensor(obj):
    """Convert an object to a dictionary of tensors."""
    snapshot = {}

    for key in dir(obj):
        # Skip dunder/hidden methods and excepted tensors
        if key.startswith("_"):
            continue

        value = getattr(obj, key, None)

        # Skip methods and functions
        if callable(value):
            continue

        # if a dict is passed
        if isinstance(value, dict):
            # If the value is a dict, we recursively convert it to a tensor
            snapshot[key] = object_to_tensor(value)
            continue

        # Skip byte strings
        if isinstance(value, bytes):
            continue

        # Convert the value to a tensor
        snapshot[key] = _to_tensor(key, value)

    return snapshot


def _to_tensor(key, val):
    if key in STATIC:
        return val

    if not isinstance(val, CONVERT_TO_KERAS_TYPES):
        return val

    if val is None:
        return None
    # Recursively handle dicts
    if isinstance(val, dict):
        return {k: _to_tensor(k, v) for k, v in val.items()}
    # Use float precision for all floats (including np.float32/64)
    if isinstance(val, float) or (isinstance(val, np.ndarray) and np.issubdtype(val.dtype, float)):
        dtype = BASE_FLOAT_PRECISION
    # Use int precision for all ints (including np.int32/64)
    elif isinstance(val, bool) or (isinstance(val, np.ndarray) and np.issubdtype(val.dtype, bool)):
        dtype = bool
    elif isinstance(val, int) or (isinstance(val, np.ndarray) and np.issubdtype(val.dtype, int)):
        dtype = BASE_INT_PRECISION
    else:
        dtype = None
    return keras.ops.convert_to_tensor(val, dtype=dtype)


class ZEAEncoderJSON(json.JSONEncoder):
    """
    A custom JSONEncoder that:
      - Converts NumPy arrays to native Python types.
      - Converts zea Enums to their values
    """

    def default(self, o):
        """Convert objects to JSON serializable types."""
        obj = o
        # Convert zea Enums to their values
        if isinstance(obj, enum.Enum):
            return obj.value

        # Convert NumPy types to native Python
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()

        return super().default(obj)


class ZEADecoderJSON(json.JSONDecoder):
    """
    A custom JSONDecoder that:
      - Converts lists into NumPy arrays.
      - Restores zea enum fields to their respective enum members.
    """

    # Create maps for quick enum lookups based on their .value
    _DATA_TYPES_MAP = {dt.value: dt for dt in DataTypes}
    _MOD_TYPES_MAP = {mt.value: mt for mt in ModTypes if mt.value is not None}

    def __init__(self, *args, **kwargs):
        # We supply our custom object_hook
        super().__init__(object_hook=self._object_hook, *args, **kwargs)

    def _object_hook(self, obj):
        """
        Called once for every JSON object (dict). We iterate through each key/value
        to see if we need to convert it into an enum or a NumPy array.
        """
        for key, value in list(obj.items()):
            # Convert lists to NumPy arrays
            if isinstance(value, list):
                # If you want a more selective approach (e.g. only numeric lists -> arrays),
                # you could check if all elements are numeric before converting.
                obj[key] = np.array(value)

            # Convert strings to DataTypes enum if it matches
            elif isinstance(value, str) and value in self._DATA_TYPES_MAP:
                obj[key] = self._DATA_TYPES_MAP[value]

            # Convert string to ModTypes enum if it matches. Also, allow None for the 'modtype' key.
            elif (key == "modtype" and value is None) or (
                isinstance(value, str) and value in self._MOD_TYPES_MAP
            ):
                obj[key] = self._MOD_TYPES_MAP[value] if value is not None else None

        return obj



"""Create a new config file by asking the user for input."""

import sys
from pathlib import Path

import schema

from zea.config import Config
from zea.internal.config.parameters import PARAMETER_DESCRIPTIONS
from zea.internal.config.validation import check_config, config_schema
from zea.log import green, red
from zea.utils import get_date_string, strtobool


def _get_input_value(config, schema_key, schema_value, descriptions):
    while True:
        input_val = input(f"Enter a value for {schema_key}: ")
        if not isinstance(schema_key, str):
            _key = schema_key.key
        else:
            _key = schema_key
        if input_val == "help":
            if _key not in descriptions:
                print(red(f"No description available for {_key}"))
                continue
            print("\t" + green(descriptions[_key]))
            continue
        try:
            config[_key] = input_val
            if isinstance(schema_value, schema.And):
                for _type in schema_value.args:
                    try:
                        config[_key] = _type(config[_key])
                        break
                    except Exception:
                        pass
                schema_value.validate(config[_key])
            else:
                schema_value(config[_key])
            break
        except Exception as e:
            print(f"Invalid input: {red(e)}")
    return config


def create_config():
    """Create a new config file by asking the user for input."""

    def _ask_user_input(config, schema_obj, descriptions):
        for key, value in schema_obj.schema.items():
            if isinstance(value, schema.Schema):
                if not isinstance(key, str):
                    _key = key.key
                else:
                    _key = key
                if isinstance(key, schema.Optional):
                    # skip optional keys
                    continue
                config[_key] = _ask_user_input(
                    config.setdefault(_key, {}), value, descriptions[_key]
                )
            elif not isinstance(key, schema.Optional):
                config = _get_input_value(config, key, value, descriptions)

        return config

    config = {}
    _ask_user_input(config, config_schema, PARAMETER_DESCRIPTIONS)

    # Ask user if they want to change any optional keys
    while True:
        try:
            key = None
            input_val = input("Do you want to change any optional keys? (yes/no): ")
            change_optional = strtobool(input_val)

            if change_optional:
                key = input("Enter the key name (e.g., 'model/beamformer/param'): ")
                keys = key.split("/")
                base_schemas = [
                    "data",
                    "plot",
                    "model",
                    "preprocess",
                    "postprocess",
                    "scan",
                ]

                if len(keys) > 1:
                    if keys[0] not in base_schemas:
                        print(red(f"Invalid key {key}, please try again."))
                        continue

                if len(keys) == 1:
                    if keys[0] in base_schemas:
                        print(
                            red(
                                f"Invalid key, cannot be part of base keys {base_schemas} "
                                "please try again."
                            )
                        )
                        continue

                nested_dict = config
                for k in keys[:-1]:
                    nested_dict = nested_dict.setdefault(k, {})

                # retrieve schema value from the nested key
                schema_obj = config_schema
                for k in keys:
                    sub_keys = [
                        s.key if not isinstance(s, str) else s for s in schema_obj.schema.keys()
                    ]

                    schema_key = list(schema_obj.schema.keys())[sub_keys.index(k)]

                    schema_obj = schema_obj.schema[schema_key]

                descriptions = PARAMETER_DESCRIPTIONS
                for k in keys[:-1]:
                    descriptions = descriptions[k]

                nested_dict = _get_input_value(nested_dict, keys[-1], schema_obj, descriptions)
            else:
                print("No optional keys will be changed.")
                break
        except KeyboardInterrupt:
            print(red("KeyboardInterrupt, exiting."))
            sys.exit()
        except Exception:
            if key is None:
                print(red("Invalid input, please try again."))
            else:
                print(red(f"Invalid key: {key}, please try again."))
            continue

    return config


if __name__ == "__main__":
    print(
        f"Let's create a new config file 🪄\n"
        f"You can always type {green('help')} "
        "to get a description of the parameter."
    )
    config = create_config()
    print(config)

    config = check_config(config)

    # Save the config to a YAML file
    name = input("Enter a name for the config: ")
    timestamp = get_date_string()

    custom_configs_folder = Path("custom_configs")
    custom_configs_folder.mkdir(exist_ok=True)
    filename = custom_configs_folder / f"{timestamp}_{name}.yaml"

    Config(config).save_to_yaml(filename)

    print(f"Find your config at {str(filename)}")



"""Data subpackage for working with the ``zea`` data format.

This subpackage provides core classes and utilities for working with the zea data format,
including file and dataset access, validation, and data loading. For more information on the
``zea`` data format, see :doc:`../data-acquisition`.

Main classes
------------

- :class:`zea.data.File` -- Open and access a single zea HDF5 data file.
- :class:`zea.data.Dataset` -- Manage and iterate over a collection of zea data files.

See the data notebook for a more detailed example: :doc:`../notebooks/data/zea_data_example`

Examples usage
^^^^^^^^^^^^^^

.. code-block:: python

    from zea import File, Dataset

    # Open a single zea data file
    with File("path/to/file.hdf5", mode="r") as file:
        file.summary()
        data = file.load_data("raw_data", indices=[0])
        scan = file.scan()
        probe = file.probe()

    # Work with a dataset (folder or list of files)
    dataset = Dataset("path/to/folder", key="raw_data")
    for file in dataset:
        print(file)
    dataset.close()

Subpackage layout
-----------------

- ``file.py``: Implements :class:`zea.File` and related file utilities.
- ``datasets.py``: Implements :class:`zea.Dataset` and folder management.
- ``dataloader.py``: Data loading utilities for batching and shuffling.
- ``data_format.py``: Data validation and example dataset generation.
- ``convert/``: Data conversion tools (e.g., from external formats).

"""  # noqa: E501

from .convert.camus import sitk_load
from .data_format import (
    DatasetElement,
    generate_example_dataset,
    generate_zea_dataset,
    validate_input_data,
)
from .dataloader import H5Generator
from .datasets import Dataset, Folder
from .file import File, load_file



"""Caching utilities for function outputs.

>[!TIP]
> Caching works best for functions that take long, but output small results. If loading of a large
> cached tensor for instance take longer than the function itself,
> it is better to not cache the result.

>[!NOTE]
> It can be useful to inherit custom classes from `zea.core.Object`, as
> these classes will be serialized properly, just like regular python objects. Otherwise
> custom classes will not be recognized as equal if they have the same attributes by the
> caching mechanism.

>[!NOTE]
> For large experiments, it can be recommended to either set a custom cache directory
> or disable the cache completely. This can be done by setting the environment variable
> `ZEA_CACHE_DIR` to a custom directory or `ZEA_DISABLE_CACHE` to `1` or `true`.
> Otherwise, the cache will be stored in `~/.zea_cache` by default, which can pile up over time.

"""

import ast
import atexit
import hashlib
import inspect
import os
import pickle
import tempfile
import textwrap
from pathlib import Path

import joblib
import keras

from zea import log

_DEFAULT_ZEA_CACHE_DIR = Path.home() / ".cache" / "zea"


def _disable_cache():
    """Disable caching by creating a temporary directory and setting the environment variable."""
    os.environ["ZEA_DISABLE_CACHE"] = "1"
    _tmp_dir = tempfile.TemporaryDirectory(prefix="zea_cache_")
    atexit.register(lambda: _tmp_dir.cleanup())
    return Path(_tmp_dir.name)


def is_cache_disabled():
    """Check if caching is disabled via environment variable."""
    val = os.environ.get("ZEA_DISABLE_CACHE", "0").strip().lower()
    return val in ("1", "true", "yes")


def _make_cache_dir(path: Path):
    """Try to create the cache directory.
    If it fails, disable the cache and return a temporary directory instead."""
    try:
        path.mkdir(parents=True, exist_ok=True)
        return path
    except Exception as e:
        log.warning(
            f"Could not create cache directory {ZEA_CACHE_DIR}: {e} \n"
            + "Disabling cache globally. Set ZEA_CACHE_DIR to a different directory "
            + "to enable caching again."
        )
        return _disable_cache()


ZEA_CACHE_DIR = Path(os.environ.get("ZEA_CACHE_DIR", _DEFAULT_ZEA_CACHE_DIR)).resolve()

# Even if we cannot create the cache directory, we still want to use a temporary directory
# to avoid errors in the rest of the code (particularly huggingface)
if is_cache_disabled():
    ZEA_CACHE_DIR = _disable_cache()
else:
    ZEA_CACHE_DIR = _make_cache_dir(ZEA_CACHE_DIR)


_CACHE_DIR = ZEA_CACHE_DIR / "cached_funcs"
_CACHE_DIR.mkdir(parents=True, exist_ok=True)


def serialize_elements(key_elements: list):
    """Serialize elements of a list to generate a cache key.

    In general uses the string representation of the elements unless
    the element has a `serialized` attribute, in which case it uses that.
    For instance this is useful for custom classes that inherit from `zea.core.Object`.

    Args:
        key_elements (list): List of elements to serialize. Can be nested lists
            or tuples. In this case the elements are serialized recursively.

    Returns:
        list[str]: List of serialized elements. In cases where the elements were
            lists of tuples those are combined into a single string.

    """
    serialized_elements = []
    for element in key_elements:
        if isinstance(element, (list, tuple)):
            # If element is a list or tuple, serialize its elements recursively
            element = serialize_elements(element)
            serialized_elements.append("_".join(element))
        elif hasattr(element, "serialized"):
            # Use the serialized attribute if it exists (e.g. for zea.core.Object)
            serialized_elements.append(str(element.serialized))
        elif isinstance(element, str):
            # If element is a string, use it as is
            serialized_elements.append(element)
        elif isinstance(element, keras.random.SeedGenerator):
            # If element is a SeedGenerator, use the state
            element = keras.ops.convert_to_numpy(element.state.value)
            element = pickle.dumps(element)
            element = hashlib.md5(element).hexdigest()
            serialized_elements.append(element)
        else:
            # Otherwise, serialize the element using pickle and hash it
            element = pickle.dumps(element)
            element = hashlib.md5(element).hexdigest()
            serialized_elements.append(element)

    return serialized_elements


def get_function_source(func):
    """Recursively get the source code of a function and its nested functions."""
    try:
        source = inspect.getsource(func)
    except OSError:
        return None  # Do not cache if source code cannot be retrieved

    # Parse the source code into an AST
    source = textwrap.dedent(source)
    tree = ast.parse(source)
    called_functions = set()

    class FunctionCallVisitor(ast.NodeVisitor):
        """AST visitor to collect function calls."""

        def visit_Call(self, node):
            """Visit a Call node and add the function name to the set."""
            if isinstance(node.func, ast.Name):
                called_functions.add(node.func.id)
            self.generic_visit(node)

    FunctionCallVisitor().visit(tree)

    # Sorting the called functions to ensure consistent cache keys
    for called_func_name in sorted(called_functions):
        try:
            called_func = func.__globals__.get(called_func_name)
            if inspect.isfunction(called_func) and called_func.__module__ != "zea.internal.cache":
                nested_source = get_function_source(called_func)
                if nested_source is None:
                    # If any nested function's source cannot be retrieved, do not cache
                    return None
                source += nested_source
        except (NameError, TypeError):
            continue

    return source


def generate_cache_key(func, args, kwargs, arg_names):
    """Generate a unique cache key based on function name and specified parameters."""
    key_elements = [func.__qualname__]  # qualified function name
    source = get_function_source(func)
    if source is None:
        log.warning(
            f"Could not get source code for function {func.__qualname__}. Not caching the result."
        )
        return None  # Do not cache if source code cannot be retrieved
    key_elements.append(source)  # source code
    if not arg_names:
        key_elements.extend(args)
        key_elements.extend(v for _, v in sorted(kwargs.items()))
    else:
        sig = inspect.signature(func)
        bound_args = sig.bind_partial(*args, **kwargs)
        for name in arg_names:
            if name in bound_args.arguments:
                key_elements.append(bound_args.arguments[name])

    key = "_".join(serialize_elements(key_elements))
    return f"{func.__qualname__}_" + hashlib.md5(key.encode()).hexdigest()


def cache_output(*arg_names, verbose=False):
    """Decorator to cache function outputs using joblib."""
    assert all(isinstance(arg_name, str) for arg_name in arg_names), (
        "All argument names must be strings, "
        "please use cache_output with strings as arguments or leave it empty "
        "to cache all arguments."
    )

    def decorator(func):
        def wrapper(*args, **kwargs):
            if is_cache_disabled():
                if verbose:
                    log.info(f"Caching is globally disabled for {func.__qualname__}.")
                return func(*args, **kwargs)
            try:
                cache_key = generate_cache_key(func, args, kwargs, arg_names)
            except Exception as e:
                log.warning(
                    f"Could not cache result for {func.__qualname__}: {e}. "
                    "Running the function without caching. "
                    "Often happens for a function wrapped with jax.jit or tf.function."
                )
                return func(*args, **kwargs)
            if cache_key is None:
                return func(*args, **kwargs)  # Run function without caching
            cache_file = _CACHE_DIR / f"{cache_key}.pkl"
            if cache_file.exists():
                if verbose:
                    log.info(f"Loading cached result for {func.__qualname__}.")
                return joblib.load(cache_file)
            elif verbose:
                log.info(f"Running {func.__qualname__} and caching the result to {cache_file}.")
            result = func(*args, **kwargs)
            joblib.dump(result, cache_file)
            return result

        return wrapper

    return decorator


def clear_cache(func_name=None):
    """Clear cache files.

    If func_name is specified, only clear cache files related to that function.
    Otherwise, clear the entire cache directory.

    Provides a summary of how much was cleared and logs the information.
    """
    total_cleared = 0

    if func_name:
        pattern = f"{func_name}_*.pkl"
    else:
        pattern = "*.pkl"

    for cache_file in _CACHE_DIR.glob(pattern):
        file_size = cache_file.stat().st_size
        cache_file.unlink()
        total_cleared += file_size

    if total_cleared > 0:
        if func_name:
            log.info(
                f"Cleared {total_cleared / (1024 * 1024):.2f} MB "
                f"from cache for function '{func_name}'."
            )
        else:
            log.info(f"Cleared {log.yellow(f'{total_cleared / (1024 * 1024):.2f}')} MB from cache.")
    else:
        log.info("No cache files to clear.")


def cache_summary():
    """Print a summary of the cache, grouping by function name and summing the sizes."""
    summary = {}
    for cache_file in _CACHE_DIR.glob("*.pkl"):
        # Assuming cache files are named as '{func_name}_{hash}.pkl'
        func_name = "_".join(cache_file.stem.split("_")[:-1])
        file_size = cache_file.stat().st_size
        summary[func_name] = summary.get(func_name, 0) + file_size

    if not summary:
        log.info(f"zea cache at {_CACHE_DIR} is empty.")
        return

    log.info(f"zea cache summary at {_CACHE_DIR}:")
    for func_name, total_size in summary.items():
        log.info(
            f"Function '{func_name}' has a total cache size of {total_size / (1024 * 1024):.2f} MB"
        )



"""Agent subpackage for closing action-perception loop in ultrasound imaging.

The `agent` subpackage provides tools and utilities for agent-based algorithms within the ``zea`` framework, including mask generation and action selection strategies. See :mod:`zea.agent.masks` and :mod:`zea.agent.selection` for key functions implementing intelligent focused transmit selection, such as the :class:`zea.agent.selection.GreedyEntropy` algorithm.

For a practical example, see :doc:`../notebooks/agent/agent_example`.

Example usage
^^^^^^^^^^^^^

.. code-block:: python

    import zea
    import numpy as np

    agent = zea.agent.selection.GreedyEntropy(
        n_actions=7,
        n_possible_actions=112,
        img_width=112,
        img_height=112,
    )

    # (batch, samples, height, width)
    particles = np.random.rand(1, 10, 112, 112)
    lines, mask = agent.sample(particles)
"""

from . import masks, selection



"""Registration module for registering classes and their names to be able to
refer to them by name in config files. The module contains a decorator class
for registering classes that can be used to register a name and optionally
additional values to keys for the class.

Usage:
- In the file defining a base class and possibly the subclasses, import the
RegisterDecorator class and create a decorator object. In the items_to_register
argument, pass a list of strings that will be used as keys for the additional
keys to register values for for every registered item.
- For each subclass, decorate the class with the decorator object and pass the
name to register to the class as the first argument and optionally additional
values to register to the keys for the class as keyword arguments.
- In other code that needs to use these classes import only the registry object
and use the registry object to get the class corresponding to a name.

Example:
```datasets.py
dataset_registry = RegisterDecorator(items_to_register=['probe_used', 'scan_class'])

@dataset_registry(name='picmus', probe_used='L11-5V', scan_class=PicmusScan)
class PICMUS(Dataset):
    ...
```

In another file:
```other_file.py
from zea.data import dataset_registry

dataset_class = dataset_registry['picmus']
dataset = dataset_class()
```
"""


class RegisterDecorator:
    """Decorator class for registering classes.

    The docorator registers a name to the class and optionally registers
    additional values to keys for the class.
    """

    def __init__(self, items_to_register=None):
        # The registry is a dictionary mapping names to classes
        self.registry = {}

        # Register additional values to keys for the class
        # additional_registries is a dictionary mapping registry names to
        # dictionaries mapping classes to values (yeah that's a mouthful)
        self.additional_registries = {}

        if items_to_register is None:
            items_to_register = {}

        for reg in items_to_register:
            assert isinstance(reg, str), "Item to register must be a string"
            self.additional_registries[reg.lower()] = {}

    def __call__(self, name, **kwargs):
        """The decorator function.

        The name is the name to register to the class and the kwargs are the
        additional values to register to the class.

        Note: All names and keys are converted to lowercase.
        """
        assert isinstance(name, str), "Name must be a string"
        assert name not in self.registry, f"Name {name} already registered"

        call_kwargs = kwargs.copy()
        name = name.lower()

        def _register(cls, name=name):
            self.registry[name] = cls

            for regname, value in call_kwargs.items():
                # If there is an additional registry with name regname,
                # register the value to the class.
                if regname in self.additional_registries:
                    # Add the class as key In the additional registry with
                    # name regname and the value as value
                    self.additional_registries[regname][cls] = value

            return cls

        return _register

    def get_parameter(self, cls_or_name, parameter):
        """Get parameter.

        Returns the value of the parameter for the class with the given
        class or name. This value can be a string or a class type.
        """
        if isinstance(cls_or_name, str):
            cls_or_name = self.registry[cls_or_name.lower()]
        # Assert that key is a class type
        assert isinstance(cls_or_name, type) or callable(cls_or_name), (
            "Key must be a class type or function"
        )
        return self.additional_registries[parameter.lower()][cls_or_name]

    def __str__(self) -> str:
        """String representation of the registry.

        Prints the keys and class names of the registry each on a single
        line followed by the keys and values of each additional registry.
        """
        string = "registry:\n"
        for key, cls in self.registry.items():
            string += f"{key.ljust(30)}: {cls.__name__}\n"

        string += "\nadditional_registries:\n"
        for reg, dictionary in self.additional_registries.items():
            string += f"{reg}:\n"
            for cls, val in dictionary.items():
                string += f"\t{cls.__name__.ljust(30)}: {val}\n"

        return string

    def __getitem__(self, key):
        """Returns the class corresponding to the key.

        The key can be a string or a class type.
        """
        assert isinstance(key, str), "Key must be a string"
        try:
            return self.registry[key.lower()]
        except KeyError as exc:
            raise KeyError(
                f"Name {key} not registered. Please choose from {self.registered_names()}."
            ) from exc

    def get_name(self, cls):
        """Retrieve the registry key name associated with the given class or instance."""
        for name, registered_class in self.registry.items():
            if isinstance(cls, type):  # cls is a class
                if issubclass(cls, registered_class):
                    return name
            else:  # cls is an instance
                if isinstance(cls, registered_class):
                    return name
        raise KeyError(f"Class or instance {cls} not registered.")

    def get_additional_registries(self):
        """Returns a list of the names of the additional registries."""
        return list(self.additional_registries.keys())

    def registered_names(self):
        """Returns a list of the names registered."""
        return list(self.registry.keys())

    def __contains__(self, key):
        """Returns True if the key is registered."""
        return key.lower() in self.registry

    def __iter__(self):
        """Returns an iterator over the keys of the registry."""
        return iter(self.registry)

    def clear(self):
        """Clears the registry."""
        self.registry = {}
        self.additional_registries = {}
        items_to_register = self.additional_registries.keys()

        if items_to_register is None:
            items_to_register = {}

        for reg in items_to_register:
            self.additional_registries[reg.lower()] = {}

    def filter_by_argument(self, argument, value):
        """Filter the registry by the given argument and value.

        Returns a list of names of classes that have the given value for the
        given argument.
        """
        return [
            name
            for name, cls in self.registry.items()
            if self.get_parameter(cls, argument) == value
        ]


# The registry for the probes.
probe_registry = RegisterDecorator()

# The registry for the beamformers.
# separate registries to avoid dupicate entries
# as torch and tf beamformers cannot coexist in same registry
tf_beamformer_registry = RegisterDecorator(items_to_register=["name", "framework"])

torch_beamformer_registry = RegisterDecorator(items_to_register=["name", "framework"])

metrics_registry = RegisterDecorator(items_to_register=["name", "framework", "supervised"])

checks_registry = RegisterDecorator(items_to_register=["data_type"])
ops_registry = RegisterDecorator(items_to_register=["name"])
ops_dep_registry = RegisterDecorator(items_to_register=["name"])
model_registry = RegisterDecorator(items_to_register=["name"])
diffusion_guidance_registry = RegisterDecorator(items_to_register=["name"])
operator_registry = RegisterDecorator(items_to_register=["name"])
action_selection_registry = RegisterDecorator(items_to_register=["name"])



"""Augmentation layers for ultrasound data."""

import keras
import numpy as np
from keras import layers, ops

from zea.tensor_ops import is_jax_prng_key, split_seed


class RandomCircleInclusion(layers.Layer):
    """
    Adds a circular inclusion to the image, optionally at random locations.

    Since this can accept N-dimensional inputs, you'll need to specify your
    ``circle_axes`` -- these are the axes onto which a circle will be drawn.
    This circle will then be broadcast along the remaining dimensions.

    You can then optionally specify whether there is a batch dim,
    and whether the circles should be located randomly across that batch.

    For example, if you have a batch of videos, e.g. of shape [batch, frame, height, width],
    then you might want to specify ``circle_axes=(2, 3)``, and
    ``randomize_location_across_batch=True``. This would result in a circle that is located
    in the same place per video, but different locations for different videos.

    Once your method has recovered the circles, you can evaluate them using
    the ``evaluate_recovered_circle_accuracy()`` method, which will expect an input
    shape matching your inputs to ``call()``.
    """

    def __init__(
        self,
        radius: int,
        fill_value: float = 1.0,
        circle_axes: tuple[int, int] = (1, 2),
        with_batch_dim=True,
        return_centers=False,
        recovery_threshold=0.1,
        randomize_location_across_batch=True,
        seed=None,
        **kwargs,
    ):
        """
        Initialize RandomCircleInclusion.

        Args:
            radius (int): Radius of the circle to include.
            fill_value (float): Value to fill inside the circle.
            circle_axes (tuple[int, int]): Axes along which to draw the circle (height, width).
            with_batch_dim (bool): Whether input has a batch dimension.
            return_centers (bool): Whether to return circle centers along with images.
            recovery_threshold (float): Threshold for considering a pixel as recovered.
            randomize_location_across_batch (bool): If True, randomize circle location
                per batch element.
            seed (Any): Optional random seed for reproducibility.
            **kwargs: Additional keyword arguments for the parent Layer.
        """
        super().__init__(**kwargs)
        self.radius = radius
        self.fill_value = fill_value
        self.circle_axes = circle_axes
        self.with_batch_dim = with_batch_dim
        self.return_centers = return_centers
        self.recovery_threshold = recovery_threshold
        self.randomize_location_across_batch = randomize_location_across_batch
        self.seed = seed
        self._axis1 = None
        self._axis2 = None
        self._perm = None
        self._inv_perm = None
        self._static_shape = None
        self._static_batch = None
        self._static_h = None
        self._static_w = None
        self._static_flat_batch = 1

    def build(self, input_shape):
        """
        Build the layer and compute static shape and permutation info.

        Args:
            input_shape (tuple): Shape of the input tensor.
        """
        rank = len(input_shape) - 1 if self.with_batch_dim else len(input_shape)
        a1, a2 = self.circle_axes
        if self.with_batch_dim and (a1 == 0 or a2 == 0):
            raise ValueError("The circle axes should not be a batch dim")
        if a1 < 0:
            a1 += rank
        elif a1 > 0 and self.with_batch_dim:
            a1 -= 1
        if a2 < 0:
            a2 += rank
        elif a2 > 0 and self.with_batch_dim:
            a2 -= 1
        if not (0 <= a1 < rank and 0 <= a2 < rank):
            raise ValueError(f"circle_axes {self.circle_axes} out of range for rank {rank}")
        if a1 == a2:
            raise ValueError("circle_axes must be two distinct axes")
        self._axis1, self._axis2 = a1, a2

        all_axes = list(range(rank))
        other_axes = [ax for ax in all_axes if ax not in (a1, a2)]
        self._perm = other_axes + [a1, a2]
        inv = [0] * rank
        for i, ax in enumerate(self._perm):
            inv[ax] = i
        self._inv_perm = inv

        if self.with_batch_dim:
            input_shape = input_shape[1:]  # ignore batch dim
        permuted_shape = [input_shape[ax] for ax in self._perm]
        if len(permuted_shape) > 2:
            self._static_flat_batch = int(np.prod(permuted_shape[:-2]))
        self._static_h = int(permuted_shape[-2])
        self._static_w = int(permuted_shape[-1])
        self._static_shape = tuple(permuted_shape)

        super().build(input_shape)

    def compute_output_shape(self, input_shape):
        """
        Compute output shape for the layer.

        Args:
            input_shape (tuple): Shape of the input tensor.

        Returns:
            tuple: The output shape (same as input).
        """
        return input_shape

    def _permute_axes_to_circle_last(self, x):
        """
        Permute axes so that circle axes are last.

        Args:
            x (Tensor): Input tensor.

        Returns:
            Tensor: Tensor with circle axes as the last two dimensions.
        """
        return ops.transpose(x, axes=self._perm)

    def _flatten_batch_and_other_dims(self, x):
        """
        Flatten all axes except the last two (circle axes).

        Args:
            x (Tensor): Input tensor with circle axes last.

        Returns:
            tuple: (reshaped tensor, flat batch size, height, width).
        """
        shape = x.shape
        flat_batch = int(np.prod(shape[:-2])) if len(shape) > 2 else 1
        h, w = shape[-2], shape[-1]
        return ops.reshape(x, [flat_batch, h, w]), flat_batch, h, w

    def _make_circle_mask(self, centers, h, w, radius, dtype):
        """
        Create a mask for each center (batch, h, w) using Keras ops.

        Args:
            centers (Tensor): Tensor of shape (batch, 2) with circle centers.
            h (int): Height of the image.
            w (int): Width of the image.
            radius (int): Radius of the circle.
            dtype (str or dtype): Data type for the mask.

        Returns:
            Tensor: Mask of shape (batch, h, w).
        """
        Y = ops.arange(h)
        X = ops.arange(w)
        Y, X = ops.meshgrid(Y, X, indexing="ij")
        Y = ops.expand_dims(Y, 0)  # (1, h, w)
        X = ops.expand_dims(X, 0)  # (1, h, w)
        # cx = ops.cast(centers[:, 0], "float32")[:, None, None]
        # cy = ops.cast(centers[:, 1], "float32")[:, None, None]
        cx = centers[:, 0][:, None, None]
        cy = centers[:, 1][:, None, None]
        dist2 = (X - cx) ** 2 + (Y - cy) ** 2
        mask = ops.cast(dist2 <= radius**2, dtype)
        return mask

    def call(self, x, seed=None):
        """
        Apply the random circle inclusion augmentation.

        Args:
            x (Tensor): Input tensor.
            seed (Any, optional): Optional random seed for reproducibility.

        Returns:
            Tensor or tuple: Augmented images, and optionally the circle
                centers if return_centers is True.
        """
        if keras.backend.backend() == "jax" and not is_jax_prng_key(seed):
            raise NotImplementedError(
                "jax.random.key() is not supported, please use jax.random.PRNGKey()"
            )
        seed = seed if seed is not None else self.seed

        if self.with_batch_dim:
            x_is_symbolic_tensor = not isinstance(ops.shape(x)[0], int)
            if x_is_symbolic_tensor:
                if self.randomize_location_across_batch:
                    imgs, centers = ops.map(lambda arg: self._call(arg, seed), x)
                else:
                    raise NotImplementedError(
                        "You cannot fix circle locations across while using"
                        + "RandomCircleInclusion as a dataset augmentation, "
                        + "since samples in a batch are handled independently."
                    )
            else:
                if self.randomize_location_across_batch:
                    batch_size = ops.shape(x)[0]
                    seeds = split_seed(seed, batch_size)
                    if all(seed is seeds[0] for seed in seeds):
                        imgs, centers = ops.map(lambda arg: self._call(arg, seeds[0]), x)
                    else:
                        imgs, centers = ops.map(
                            lambda args: self._call(args[0], args[1]), (x, seeds)
                        )
                else:
                    imgs, centers = ops.map(lambda arg: self._call(arg, seed), x)
        else:
            imgs, centers = self._call(x, seed)

        if self.return_centers:
            return imgs, centers
        else:
            return imgs

    def _call(self, x, seed):
        """
        Internal method to apply the augmentation to a single image.

        Args:
            x (Tensor): Input image tensor with circle axes last.
            seed (Any): Random seed for circle location.

        Returns:
            tuple: (augmented image, center coordinates).
        """
        x = self._permute_axes_to_circle_last(x)
        flat, flat_batch_size, h, w = self._flatten_batch_and_other_dims(x)

        def _draw_circle_2d(img2d):
            cx = ops.cast(
                keras.random.uniform((), self.radius, w - self.radius, seed=seed),
                "int32",
            )
            new_seed, _ = split_seed(seed, 2)  # ensure that cx and cy are independent
            cy = ops.cast(
                keras.random.uniform((), self.radius, h - self.radius, seed=new_seed),
                "int32",
            )
            mask = self._make_circle_mask(
                ops.stack([cx, cy])[None, :], h, w, self.radius, img2d.dtype
            )[0]
            img_aug = img2d * (1 - mask) + self.fill_value * mask
            center = ops.stack([cx, cy])
            return img_aug, center

        aug_imgs, centers = ops.vectorized_map(_draw_circle_2d, flat)
        aug_imgs = ops.reshape(aug_imgs, x.shape)
        aug_imgs = ops.transpose(aug_imgs, axes=self._inv_perm)
        centers_shape = [2] if flat_batch_size == 1 else [flat_batch_size, 2]
        centers = ops.reshape(centers, centers_shape)
        return (aug_imgs, centers)

    def get_config(self):
        """
        Get layer configuration for serialization.

        Returns:
            dict: Dictionary of layer configuration.
        """
        cfg = super().get_config()
        cfg.update(
            {
                "radius": self.radius,
                "fill_value": self.fill_value,
                "circle_axes": self.circle_axes,
                "return_centers": self.return_centers,
            }
        )
        return cfg

    def evaluate_recovered_circle_accuracy(
        self, images, centers, recovery_threshold, fill_value=None
    ):
        """
        Evaluate the percentage of the true circle that has been recovered in the images.

        Args:
            images (Tensor): Tensor of images (any shape, with circle axes as specified).
            centers (Tensor): Tensor of circle centers (matching batch size).
            recovery_threshold (float): Threshold for considering a pixel as recovered.
            fill_value (float, optional): Optionally override fill_value for cases
                where image range has changed.

        Returns:
            Tensor: Percentage recovered for each circle (shape: [num_circles]).
        """
        fill_value = fill_value or self.fill_value

        def _evaluate_recovered_circle_accuracy(image, center):
            image_perm = self._permute_axes_to_circle_last(image)
            h, w = image_perm.shape[-2], image_perm.shape[-1]
            flat_image, _, _, _ = self._flatten_batch_and_other_dims(image_perm)
            flat_center = ops.reshape(center, [-1, 2])
            mask = self._make_circle_mask(flat_center, h, w, self.radius, flat_image.dtype)
            diff = ops.abs(flat_image - fill_value)
            recovered = ops.cast(diff <= recovery_threshold, flat_image.dtype) * mask
            recovered_sum = ops.sum(recovered, axis=[1, 2])
            mask_sum = ops.sum(mask, axis=[1, 2])
            percent_recovered = recovered_sum / (mask_sum + 1e-8)
            return percent_recovered

        if self.with_batch_dim:
            return ops.vectorized_map(
                lambda args: _evaluate_recovered_circle_accuracy(args[0], args[1]),
                (images, centers),
            )[..., 0]
        else:
            return _evaluate_recovered_circle_accuracy(images, centers)



"""Tiny Autoencoder (TAESD) model converted to Tensorflow.

For the original implementation, see the `TAESD repository <https://github.com/madebyollin/taesd>`_.

You can see an example of how to use this model in the example notebook:
:doc:`../notebooks/models/taesd_autoencoder_example`.

"""

from pathlib import Path

import keras
from keras import backend, ops

from zea.backend import _import_tf
from zea.internal.registry import model_registry
from zea.models.base import BaseModel
from zea.models.preset_utils import get_preset_loader, register_presets
from zea.models.presets import taesdxl_decoder_presets, taesdxl_encoder_presets, taesdxl_presets

tf = _import_tf()


@model_registry(name="taesdxl")
class TinyAutoencoder(BaseModel):
    """[TAESD](https://github.com/madebyollin/taesd) model in TensorFlow."""

    def __init__(self, **kwargs):
        """
        Initializes the TAESD model with the given parameters.

        Args:
            **kwargs: Additional keyword arguments to pass to the superclass initializer.
        """
        if backend.backend() not in ["tensorflow", "jax"]:
            raise NotImplementedError(
                "TinyDecoder is only currently supported with the TensorFlow or Jax backend."
            )

        assert tf is not None, (
            "TensorFlow is not installed. Please install TensorFlow to use EchoNetDynamic. This is "
            "required even if you are using the Jax backend, the model is built using TensorFlow."
        )

        _fix_tf_to_jax_resize_nearest_neighbor()

        super().__init__(**kwargs)

        self.encoder = TinyEncoder()
        self.decoder = TinyDecoder()

        self._grayscale = False

    def encode(self, inputs):
        """Encode the input images.

        Args:
            inputs (tensor): Input images of shape (batch_size, height, width, channels).
        """
        if self.encoder.network is None or self.decoder.network is None:
            raise ValueError(
                "Please load model using `TinyAutoencoder.from_preset()` before calling."
            )

        if ops.shape(inputs)[-1] == 1:
            self._grayscale = True
            inputs = ops.concatenate([inputs, inputs, inputs], axis=-1)  # grayscale to RGB
        return self.encoder(inputs)

    def decode(self, inputs):
        """Decode the encoded images.

        Args:
            inputs (tensor): Input images of shape (batch_size, height, width, 4).
        """
        decoded = self.decoder(inputs)
        if self._grayscale:
            decoded = ops.image.rgb_to_grayscale(decoded, data_format="channels_last")
        return decoded

    def call(self, inputs):
        """Applies the full autoencoder to the input."""
        encoded = self.encode(inputs)
        # NOTE: Here you can compress the encoding a little bit more by going
        # to uint8 like in the original model
        # https://github.com/huggingface/diffusers/blob/cd30820/src/diffusers/models/autoencoders/autoencoder_tiny.py?plain=1#L336-L342 # noqa: E501
        decoded = self.decode(encoded)
        return decoded

    def custom_load_weights(self, preset, **kwargs):
        """Load the weights for the encoder and decoder."""
        self.encoder.custom_load_weights(preset)
        self.decoder.custom_load_weights(preset)


class TinyBase(BaseModel):
    """Base class for TAESD encoder and decoder."""

    def __init__(self, tiny_type=None, **kwargs):
        # Assertions
        assert tiny_type in [
            "encoder",
            "decoder",
        ], "Type must be either 'encoder' or 'decoder'."
        if backend.backend() not in ["tensorflow", "jax"]:
            raise NotImplementedError(
                f"{self.__class__.__name__} is only currently supported with the "
                "TensorFlow or Jax backend."
            )

        super().__init__(**kwargs)
        self.network = None

        self.download_files = [
            f"{tiny_type}/variables/variables.data-00000-of-00001",
            f"{tiny_type}/variables/variables.index",
            f"{tiny_type}/saved_model.pb",
            f"{tiny_type}/fingerprint.pb",
        ]

    def build(self, input_shape):
        """Builds the network."""
        self.maybe_convert_to_jax(input_shape)

    def maybe_convert_to_jax(self, input_shape):
        """Converts the network to Jax if backend is Jax."""
        if backend.backend() == "jax":
            inputs = ops.zeros(input_shape)
            from zea.backend import tf2jax

            jax_func, jax_params = tf2jax.convert(tf.function(self.network), inputs)

            def call_fn(params, state, rng, inputs, training):
                return jax_func(state, inputs)

            self.network = keras.layers.JaxLayer(call_fn, state=jax_params)

    def _load_layer(self, path: Path | str):
        if backend.backend() == "tensorflow":
            return keras.layers.TFSMLayer(path, call_endpoint="serving_default")
        elif backend.backend() == "jax":
            return tf.saved_model.load(path)
        else:
            raise NotImplementedError(
                f"{self.__class__.__name__} is only currently supported with the "
                f"TensorFlow or Jax backend. You are using {backend.backend()}."
            )

    def custom_load_weights(self, preset, **kwargs):
        """Load the weights for the encoder or decoder."""
        loader = get_preset_loader(preset)

        for file in self.download_files:
            filename = loader.get_file(file)

        base_path = Path(filename).parent
        self.network = self._load_layer(base_path)

    def call(self, inputs):
        """
        Applies the network to the input.
        """
        if self.network is None:
            raise ValueError(
                f"Please load model using `{self.__class__.__name__}.from_preset()` before calling."
            )

        out = self.network(inputs)
        if backend.backend() == "tensorflow":
            # because decoded is dict, take first key
            out = out[next(iter(out))]
        return out


@model_registry(name="taesdxl_encoder")
class TinyEncoder(TinyBase):
    """Encoder from TAESD model."""

    def __init__(self, **kwargs):
        """
        Initializes the TAESD encoder.

        Args:
            **kwargs: Additional keyword arguments passed to the superclass initializer.
        """
        super().__init__(tiny_type="encoder", **kwargs)


@model_registry(name="taesdxl_decoder")
class TinyDecoder(TinyBase):
    """Decoder from TAESD model."""

    def __init__(self, **kwargs):
        """
        Initializes the TAESD decoder.

        Args:
            **kwargs: Additional keyword arguments passed to the superclass initializer.
        """
        super().__init__(tiny_type="decoder", **kwargs)


def _fix_tf_to_jax_resize_nearest_neighbor():
    # This block of code is used to allow the Jax backend to work with TAESD
    # It overrides the ResizeNearestNeighbor op to allow align_corners=True
    # and half_pixel_centers=True. This means outputs of the jax model might
    # not be a 100% match to the tensorflow model
    if backend.backend() != "jax":
        return

    import jax
    import jax.numpy as jnp

    from zea.backend import tf2jax

    def _resize_nearest_neighbor(proto):
        """Parse a ResizeNearestNeighbor op."""
        tf2jax._src.ops._check_attrs(proto, {"T", "align_corners", "half_pixel_centers"})

        def _func(images: jnp.ndarray, size: jnp.ndarray) -> jnp.ndarray:
            if len(images.shape) != 4:
                raise ValueError(
                    "Expected A 4D tensor with shape [batch, height, width, channels], "
                    f"found {images.shape}"
                )

            inp_batch, _, _, inp_channels = images.shape
            out_height, out_width = size.tolist()

            return jax.image.resize(
                images,
                shape=(inp_batch, out_height, out_width, inp_channels),
                method=jax.image.ResizeMethod.NEAREST,
            )

        return _func

    # hack to allow align_corners=True and half_pixel_centers=True
    tf2jax._src.ops._jax_ops["ResizeNearestNeighbor"] = _resize_nearest_neighbor


register_presets(taesdxl_presets, TinyAutoencoder)
register_presets(taesdxl_encoder_presets, TinyEncoder)
register_presets(taesdxl_decoder_presets, TinyDecoder)



"""Huggingface hub (hf) tooling."""

from pathlib import Path, PurePosixPath

from huggingface_hub import HfApi, login, snapshot_download

from zea import log
from zea.data.preset_utils import _hf_list_files, _hf_parse_path

HF_PREFIX = "hf://"


def load_model_from_hf(repo_id, revision="main", verbose=True):
    """
    Load the model from a given repo_id using the Hugging Face library.

    Will download to a `model_dir` directory and return the path to it.
    Need your own load model logic to load the model from the `model_dir`.

    Args:
        repo_id (str): The ID of the repository.
        revision (str): The revision to download. Can be a branch, tag, or commit hash.
        verbose (bool): Whether to print the download message. Default is True.

    Returns:
        model_dir (Path): The path to the downloaded model directory.

    """
    login(new_session=False)

    model_dir = snapshot_download(
        repo_id=repo_id,
        repo_type="model",
        revision=revision,
    )
    api = HfApi()
    commit = api.list_repo_commits(repo_id, revision=revision)[0]
    commit_message = commit.title
    commit_time = commit.created_at.strftime("%B %d, %Y at %I:%M %p %Z")

    if verbose:
        log.info(
            log.yellow(
                f"Succesfully loaded model {commit_message} from "
                f"'https://huggingface.co/{repo_id}'. Last updated on {commit_time}."
            )
        )

    return Path(model_dir)


def upload_folder_to_hf(
    local_dir,
    repo_id,
    commit_message=None,
    revision="main",
    tag=None,
    verbose=True,
):
    """
    Upload a local directory to Hugging Face Hub.

    Args:
        local_dir (str or Path): Path to the local directory to upload.
        repo_id (str): The ID of the repository to upload to.
        commit_message (str, optional): Commit message. Defaults to "Upload files".
        revision (str): The revision to upload to. Defaults to "main".
        tag (str, optional): Tag to create. Defaults to None.
        verbose (bool): Whether to print the upload message. Default is True.

    Returns:
        str: URL of the uploaded repository.
    """
    login(new_session=False)
    api = HfApi()

    local_dir = Path(local_dir)
    if not commit_message:
        commit_message = f"Upload files from {local_dir.name}"

    # create branch if it doesn't exist
    api.create_branch(repo_id, repo_type="model", branch=revision, exist_ok=True)

    api.upload_folder(
        folder_path=local_dir,
        repo_id=repo_id,
        repo_type="model",
        commit_message=commit_message,
        revision=revision,
    )

    if tag:
        api.create_tag(repo_id, repo_type="model", tag=tag)

    if verbose:
        msg = f"Uploaded files from '{local_dir}' to 'https://huggingface.co/{repo_id}'."
        if tag:
            msg += f" Tagged as {tag}."
        log.info(log.yellow(msg))

    return f"https://huggingface.co/{repo_id}"


class HFPath(PurePosixPath):
    """A path-like object that preserves the hf:// scheme and mimics Path API."""

    _scheme = HF_PREFIX

    def __new__(cls, *args):
        # Strip "hf://" from all arguments and normalize
        parts = []
        for arg in args:
            s = str(arg)
            if s.startswith(cls._scheme):
                s = s[len(cls._scheme) :]
            parts.append(s.strip("/"))
        combined = "/".join(parts)
        # Store path without scheme
        self = super().__new__(cls, combined)
        # Mark this as an HF path that needs a scheme when stringified
        self._needs_scheme = True
        return self

    def __str__(self):
        # Get the raw path string without any scheme
        path_str = PurePosixPath.__str__(self)

        # Remove any hf:/ prefix if it somehow got included
        if path_str.startswith("hf:/"):
            path_str = path_str[len("hf:/") :]

        # Add our scheme prefix if this is meant to be an HF path
        if getattr(self, "_needs_scheme", True):
            return f"{self._scheme}{path_str}"
        return path_str

    def __truediv__(self, key):
        return self.__class__(self, key)

    def joinpath(self, *args):
        """Join paths like Path.joinpath but preserve the hf:// scheme."""
        return self.__class__(self, *args)

    @property
    def repo_id(self):
        """Extract the repo ID (e.g., zeahub/camus-sample)."""
        parts = [p for p in self.parts if p and p != "hf:"]
        if len(parts) < 2:
            raise ValueError("Invalid HFPath: cannot extract repo_id")
        return f"{parts[0]}/{parts[1]}"

    @property
    def subpath(self):
        """Get path inside the repo."""
        return "/".join(self.parts[3:])

    def is_file(self):
        """Return True if this HFPath points to a file in the repo."""
        repo_id, subpath = _hf_parse_path(str(self))
        if not subpath:
            return False
        files = _hf_list_files(repo_id)
        return any(f == subpath for f in files)

    def is_dir(self):
        """Return True if this HFPath points to a directory in the repo."""
        repo_id, subpath = _hf_parse_path(str(self))
        files = _hf_list_files(repo_id)
        # If subpath is empty, it's the repo root, which is a directory
        if not subpath:
            return True
        # If any file starts with subpath + '/', it's a directory
        prefix = subpath.rstrip("/") + "/"
        return any(f.startswith(prefix) for f in files)



"""Display functionality, including scan conversion frustrum conversion, etc."""

from functools import partial
from typing import Tuple, Union

import numpy as np
import scipy
from keras import ops
from PIL import Image

from zea import log
from zea.tools.fit_scan_cone import fit_and_crop_around_scan_cone
from zea.utils import translate


def to_8bit(image, dynamic_range: Union[None, tuple] = None, pillow: bool = True):
    """Convert image to 8 bit image [0, 255]. Clip between dynamic range.

    Args:
        image (ndarray): Input image(s). Should be in between dynamic range.
        dynamic_range (tuple, optional): Dynamic range of input image(s).
        pillow (bool, optional): Whether to return PIL image. Defaults to True.

    Returns:
        image (ndarray): Output 8 bit image(s) [0, 255].

    """
    if dynamic_range is None:
        dynamic_range = (-60, 0)

    image = ops.convert_to_numpy(image)
    image = np.clip(image, *dynamic_range)
    image = translate(image, dynamic_range, (0, 255))
    image = image.astype(np.uint8)
    if pillow:
        image = Image.fromarray(image)
    return image


def compute_scan_convert_2d_coordinates(
    image_shape,
    rho_range: Tuple[float, float],
    theta_range: Tuple[float, float],
    resolution: Union[float, None] = None,
    dtype: str = "float32",
):
    """Precompute coordinates for 2d scan conversion from polar coordinates"""
    assert len(rho_range) == 2, "rho_range should be a tuple of length 2"
    assert len(theta_range) == 2, "theta_range should be a tuple of length 2"
    assert rho_range[0] < rho_range[1], "min_rho should be less than max_rho"

    rho = ops.linspace(rho_range[0], rho_range[1], image_shape[-2], dtype=dtype)
    theta = ops.linspace(theta_range[0], theta_range[1], image_shape[-1], dtype=dtype)

    rho_grid, theta_grid = ops.meshgrid(rho, theta, indexing="ij")

    x_grid, z_grid = frustum_convert_rt2xz(rho_grid, theta_grid)

    x_lim = [ops.min(x_grid), ops.max(x_grid)]
    z_lim = [ops.min(z_grid), ops.max(z_grid)]

    d_rho = rho[1] - rho[0]
    d_theta = theta[1] - theta[0]

    if resolution is None:
        # arc length along constant phi at 1/4 depth
        sRT = 0.25 * (rho[0] + rho[-1]) * d_theta
        # average of arc lengths and radial step
        resolution = ops.mean([sRT, d_rho])  # mm per pixel

    x_vec = ops.arange(x_lim[0], x_lim[1], resolution)
    z_vec = ops.arange(z_lim[0], z_lim[1], resolution)

    z_grid, x_grid = ops.meshgrid(z_vec, x_vec)

    rho_grid_interp, theta_grid_interp = frustum_convert_xz2rt(
        x_grid, z_grid, theta_limits=[theta[0], theta[-1]]
    )

    # Map rho and theta interpolation points to grid indices
    rho_min, rho_max = ops.min(rho), ops.max(rho)
    theta_min, theta_max = ops.min(theta), ops.max(theta)
    rho_idx = (rho_grid_interp - rho_min) / (rho_max - rho_min) * (image_shape[-2] - 1)
    theta_idx = (theta_grid_interp - theta_min) / (theta_max - theta_min) * (image_shape[-1] - 1)
    # Stack coordinates as required for map_coordinates
    coordinates = ops.stack([rho_idx, theta_idx], axis=0)
    parameters = {
        "resolution": resolution,
        "x_lim": x_lim,
        "z_lim": z_lim,
        "rho_range": rho_range,
        "theta_range": theta_range,
        "d_rho": d_rho,
        "d_theta": d_theta,
    }
    return coordinates, parameters


def scan_convert_2d(
    image,
    rho_range: Tuple[float, float] = None,
    theta_range: Tuple[float, float] = None,
    resolution: Union[float, None] = None,
    coordinates: Union[None, np.ndarray] = None,
    fill_value: float = 0.0,
    order: int = 1,
    **kwargs,
):
    """
    Perform scan conversion on a 2D ultrasound image from polar coordinates
    (rho, theta) to Cartesian coordinates (x, z).

    Args:
        image (ndarray): The input 2D ultrasound image in polar coordinates.
            Has dimensions (n_rho, n_theta) with optional batch.
        rho_range (tuple): A tuple specifying the range of rho values
            (min_rho, max_rho). Defined in mm.
        theta_range (tuple): A tuple specifying the range of theta values
            (min_theta, max_theta). Defined in radians.
        resolution (float, optional): The resolution for the Cartesian grid.
            If None, it is calculated based on the input image. In mm / pixel.
        coordinates (ndarray, optional): Precomputed coordinates for scan conversion.
            If provided, it will be used instead of computing new coordinates based on
            the input image shape and ranges.
        fill_value (float, optional): The value to fill in for coordinates
            outside the input image ranges. Defaults to 0.0. When set to NaN,
            no interpolation at the edges will happen.
        order (int, optional): The order of the spline interpolation. Defaults to 1.

    Returns:
        ndarray: The scan-converted 2D ultrasound image in Cartesian coordinates.
            Has dimensions (n_z, n_x). Coordinates outside the input image
            ranges are filled with NaNs.
        parameters (dict): A dictionary containing information about the scan conversion.
            Contains the resolution, x, and z limits, rho and theta ranges.

    Note:
        Polar grid is inferred from the input image shape and the supplied
        rho and theta ranges. Cartesian grid is computed based on polar grid
        with resolutions specified by resolution parameter.

    """
    assert "float" in ops.dtype(image), "Image must be float type"

    parameters = {}
    if coordinates is None:
        coordinates, parameters = compute_scan_convert_2d_coordinates(
            image.shape, rho_range, theta_range, resolution, dtype=image.dtype
        )

    images_sc = _interpolate_batch(image, coordinates, fill_value, order=order, **kwargs)

    # swap axis to match z, x
    images_sc = ops.swapaxes(images_sc, -1, -2)

    return images_sc, parameters


def compute_scan_convert_3d_coordinates(
    image_shape,
    rho_range: Tuple[float, float],
    theta_range: Tuple[float, float],
    phi_range: Tuple[float, float],
    resolution: Union[float, None] = None,
    dtype: str = "float32",
):
    """Precompute coordinates for 3d scan conversion from polar coordinates"""
    assert len(rho_range) == 2, "rho_range should be a tuple of length 2"
    assert len(theta_range) == 2, "theta_range should be a tuple of length 2"
    assert len(phi_range) == 2, "phi_range should be a tuple of length 2"
    assert rho_range[0] < rho_range[1], "min_rho should be less than max_rho"

    rho = ops.linspace(rho_range[0], rho_range[1], image_shape[-3], dtype=dtype)
    theta = ops.linspace(theta_range[0], theta_range[1], image_shape[-2], dtype=dtype)
    phi = ops.linspace(phi_range[0], phi_range[1], image_shape[-1], dtype=dtype)

    rho_grid, theta_grid, phi_grid = ops.meshgrid(rho, theta, phi, indexing="ij")

    x_grid, y_grid, z_grid = frustum_convert_rtp2xyz(rho_grid, theta_grid, phi_grid)

    x_lim = [ops.min(x_grid), ops.max(x_grid)]
    y_lim = [ops.min(y_grid), ops.max(y_grid)]
    z_lim = [ops.min(z_grid), ops.max(z_grid)]

    d_rho = rho[1] - rho[0]
    d_theta = theta[1] - theta[0]
    d_phi = phi[1] - phi[0]

    if resolution is None:
        # arc length along constant phi at 1/4 depth
        sRT = 0.25 * (rho[0] + rho[-1]) * d_theta
        # arc length along constant theta at 1/4 depth
        sRP = 0.25 * (rho[0] + rho[-1]) * d_phi
        # average of arc lengths and radial step
        resolution = ops.mean([sRT, sRP, d_rho])  # mm per pixel

    z_vec = ops.arange(z_lim[0], z_lim[1], resolution)
    x_vec = ops.arange(x_lim[0], x_lim[1], resolution)
    y_vec = ops.arange(y_lim[0], y_lim[1], resolution)

    z_grid, x_grid, y_grid = ops.meshgrid(z_vec, x_vec, y_vec)

    rho_grid_interp, theta_grid_interp, phi_grid_interp = frustum_convert_xyz2rtp(
        x_grid,
        y_grid,
        z_grid,
        theta_limits=[theta[0], theta[-1]],
        phi_limits=[phi[0], phi[-1]],
    )

    # return volume
    rho_min, rho_max = ops.min(rho), ops.max(rho)
    theta_min, theta_max = ops.min(theta), ops.max(theta)
    phi_min, phi_max = ops.min(phi), ops.max(phi)
    rho_idx = (rho_grid_interp - rho_min) / (rho_max - rho_min) * (image_shape[-3] - 1)
    theta_idx = (theta_grid_interp - theta_min) / (theta_max - theta_min) * (image_shape[-2] - 1)
    phi_idx = (phi_grid_interp - phi_min) / (phi_max - phi_min) * (image_shape[-1] - 1)

    # Stack coordinates as required for map_coordinates
    coordinates = ops.stack([rho_idx, theta_idx, phi_idx], axis=0)
    parameters = {
        "resolution": resolution,
        "x_lim": x_lim,
        "y_lim": y_lim,
        "z_lim": z_lim,
        "rho_range": rho_range,
        "theta_range": theta_range,
        "phi_range": phi_range,
        "d_rho": d_rho,
        "d_theta": d_theta,
        "d_phi": d_phi,
    }
    return coordinates, parameters


def scan_convert_3d(
    image,
    rho_range: Tuple[float, float] = None,
    theta_range: Tuple[float, float] = None,
    phi_range: Tuple[float, float] = None,
    resolution: Union[float, None] = None,
    coordinates: Union[None, np.ndarray] = None,
    fill_value: float = 0.0,
    order: int = 1,
):
    """
    Perform scan conversion on a 3D ultrasound image from polar coordinates
    (rho, theta, phi) to Cartesian coordinates (z, x, y).

    Args:
        image (ndarray): The input 3D ultrasound image in polar coordinates.
            Has dimensions (n_rho, n_theta, n_phi) with optional batch.
        rho_range (tuple): A tuple specifying the range of rho values
            (min_rho, max_rho). Defined in mm.
        theta_range (tuple): A tuple specifying the range of theta values
            (min_theta, max_theta). Defined in radians.
        phi_range (tuple): A tuple specifying the range of phi values
            (min_phi, max_phi). Defined in radians.
        resolution (float, optional): The resolution for the Cartesian grid.
            If None, it is calculated based on the input image. In mm / pixel.
        coodinates (ndarray, optional): Precomputed coordinates for scan conversion.
            If provided, it will be used instead of computing new coordinates based on
            the input image shape and ranges.
        fill_value (float, optional): The value to fill in for coordinates
            outside the input image ranges. Defaults to 0.0. When set to NaN,
            no interpolation at the edges will happen.
        order (int, optional): The order of the spline interpolation. Defaults to 1.

    Returns:
        ndarray: The scan-converted 3D ultrasound image in Cartesian coordinates.
            Has dimensions (n_z, n_x, n_y). Coordinates outside the input image
            ranges are filled with NaNs.
        parameters (dict): A dictionary containing information about the scan conversion.
            Contains the resolution, x, y, and z limits, rho, theta, and phi ranges.

    Note:
        Polar grid is inferred from the input image shape and the supplied
        rho, theta and phi ranges. Cartesian grid is computed based on polar grid
        with resolutions specified by resolution parameter.
    """
    assert "float" in ops.dtype(image), "Image must be float type"

    parameters = {}
    if coordinates is None:
        coordinates, parameters = compute_scan_convert_3d_coordinates(
            image.shape,
            rho_range,
            theta_range,
            phi_range,
            resolution,
            dtype=image.dtype,
        )

    images_sc = _interpolate_batch(image, coordinates, fill_value, order=order)

    # swap axis to match z, x, y
    images_sc = ops.swapaxes(images_sc, -2, -3)
    return images_sc, parameters


def scan_convert(
    image,
    rho_range: Tuple[float, float] = None,
    theta_range: Tuple[float, float] = None,
    phi_range: Tuple[float, float] = None,
    resolution: Union[float, None] = None,
    coordinates: Union[None, np.ndarray] = None,
    fill_value: float = 0.0,
    order: int = 1,
    with_batch_dim: bool = False,
):
    """Scan convert image based on number of dimensions."""
    if len(image.shape) == 2 + int(with_batch_dim):
        return scan_convert_2d(
            image,
            rho_range,
            theta_range,
            resolution,
            coordinates,
            fill_value,
            order,
        )
    elif len(image.shape) == 3 + int(with_batch_dim):
        return scan_convert_3d(
            image,
            rho_range,
            theta_range,
            phi_range,
            resolution,
            coordinates,
            fill_value,
            order,
        )
    else:
        raise ValueError(
            f"Image must be 2D or 3D (with optional batch dim). Got shape: {image.shape}"
        )


def map_coordinates(inputs, coordinates, order, fill_mode="constant", fill_value=0):
    """map_coordinates using keras.ops or scipy.ndimage when order > 1."""
    if order > 1:
        inputs = ops.convert_to_numpy(inputs)
        out = scipy.ndimage.map_coordinates(
            inputs, coordinates, order=order, mode=fill_mode, cval=fill_value
        )
        return ops.convert_to_tensor(out)
    else:
        return ops.image.map_coordinates(
            inputs,
            coordinates,
            order=order,
            fill_mode=fill_mode,
            fill_value=fill_value,
        )


def _interpolate_batch(images, coordinates, fill_value=0.0, order=1, vectorize=True):
    """Interpolate a batch of images."""
    image_shape = images.shape
    num_image_dims = coordinates.shape[0]

    batch_dims = images.shape[:-num_image_dims]

    images = ops.reshape(images, (-1, *image_shape[-num_image_dims:]))

    map_coordinates_fn = partial(
        map_coordinates,
        coordinates=coordinates,
        order=order,
        fill_mode="constant",
        fill_value=fill_value,
    )

    if order > 1:
        # cpu bound
        images_sc = ops.stack(list(map(map_coordinates_fn, images)))
    elif not vectorize:
        images_sc = ops.map(map_coordinates_fn, images)
    else:
        # gpu bound
        images_sc = ops.vectorized_map(map_coordinates_fn, images)

    # ignore batch dim to get image shape
    image_sc_shape = ops.shape(images_sc)[1:]
    images_sc = ops.reshape(images_sc, (*batch_dims, *image_sc_shape))

    return images_sc


def cart2pol(x, y):
    """Convert x, y cartesian coordinates to polar coordinates theta, rho."""
    theta = ops.mod(ops.arctan2(x, -y), np.pi * 2)
    rho = ops.sqrt(x**2 + y**2)
    return (theta, rho)


def rotate_coordinates(coords, angle_deg):
    """Rotate (x, y) coordinates by a given angle in degrees."""
    angle_rad = np.deg2rad(angle_deg)
    rotation_matrix = ops.array(
        [
            [ops.cos(angle_rad), -ops.sin(angle_rad)],
            [ops.sin(angle_rad), ops.cos(angle_rad)],
        ],
        dtype=coords.dtype,
    )
    return coords @ ops.transpose(rotation_matrix)


def cartesian_to_polar_matrix(
    cartesian_matrix,
    fill_value=0.0,
    polar_shape=None,
    tip=None,
    r_max=None,
    angle=np.deg2rad(45),
    interpolation_order=1,
):
    """
    Convert a Cartesian image matrix to a polar coordinate representation.

    Args:
        cartesian_matrix (tensor): Input 2D image array in Cartesian coordinates.
        fill_value (float): Value to use for points sampled outside the input image.
        polar_shape (tuple, optional): Desired shape of the polar output (rows, cols).
            Defaults to the shape of the input image.
        tip (tuple, optional): (x, y) coordinates of the origin for the polar
            transformation (typically the probe tip). Defaults to the center-top of the image.
        r_max (float, optional): Maximum radius to consider in the polar transform.
            Defaults to the height of the input image.
        angle (float): Total angular field of view (in radians) centered at 0.
            The polar grid spans from -angle to +angle.
        interpolation_order (int): Order of interpolation to use (0 = nearest-neighbor,
            1 = linear, 2+ = spline). Matches the convention of `scipy.ndimage.map_coordinates`.

    Returns:
        polar_matrix (Array): The image re-sampled in polar coordinates with shape `polar_shape`.
    """
    if ops.dtype(cartesian_matrix) != "float32":
        log.info(
            f"Cartesian matrix with dtype {ops.dtype(cartesian_matrix)} has been cast to float32."
        )
        cartesian_matrix = ops.cast(cartesian_matrix, "float32")

    # Assume that polar grid is same shape as cartesian grid unless specified
    cartesian_rows, cartesian_cols = ops.shape(cartesian_matrix)
    if polar_shape is None:
        polar_rows, polar_cols = cartesian_rows, cartesian_cols
    else:
        polar_rows, polar_cols = polar_shape

    # assume tip is at center top unless specified
    if tip is None:
        center_x = cartesian_cols // 2
        tip_y = 0
        tip = (center_x, tip_y)

    # assume r_max is the total height of the input image unless specified
    if r_max is None:
        r_max = cartesian_rows

    center_x, center_y = tip

    # Interpolation grid in polar coordinates
    r = ops.linspace(0, r_max, polar_rows, dtype="float32")
    theta = ops.linspace(-angle, angle, polar_cols, dtype="float32")
    r_grid, theta_grid = ops.meshgrid(r, theta)

    # convert discretized radii and angle intervals to polar coordinates
    x_polar = r_grid * ops.cos(theta_grid)
    y_polar = r_grid * ops.sin(theta_grid)

    # Inverse rotation to match original orientation
    polar_coords = ops.stack([ops.ravel(x_polar), ops.ravel(y_polar)], axis=0)
    polar_coords_rotated = ops.transpose(rotate_coordinates(ops.transpose(polar_coords), 90))

    # Shift to image indices
    yq = polar_coords_rotated[1, :] + center_y
    xq = polar_coords_rotated[0, :] + center_x
    coords_for_interp = ops.stack([yq, xq])

    polar_values = map_coordinates(
        cartesian_matrix,
        coords_for_interp,
        order=interpolation_order,
        fill_mode="constant",
        fill_value=fill_value,
    )

    polar_matrix = ops.rot90(ops.reshape(polar_values, (polar_cols, polar_rows)), k=-1)
    return polar_matrix


def inverse_scan_convert_2d(
    cartesian_image,
    fill_value=0.0,
    angle=np.deg2rad(45),
    output_size=None,
    interpolation_order=1,
    find_scan_cone=True,
):
    """
    Convert a Cartesian-format ultrasound image to a polar representation.

    This function can be used to recover a sector-shaped scan (polar format)
    from a Cartesian representation of an image.
    Optionally, it can detect and crop around the scan cone before conversion.

    Args:
        cartesian_image (tensor): 2D image array in Cartesian coordinates.
        fill_value (float): Value used to fill regions outside the original image
            during interpolation.
        angle (float): Angular field of view (in radians) used for the polar transformation.
            The polar output will span from -angle to +angle.
        output_size (tuple, optional): Shape (rows, cols) of the resulting polar image.
            If None, the shape of the input image is used.
        interpolation_order (int): Order of interpolation used in resampling
            (0 = nearest-neighbor, 1 = linear, etc.).
        find_scan_cone (bool): If True, automatically detects and crops around the scan cone
            in the Cartesian image before polar conversion, ensuring that the scan cone is
            centered without padding. Can be set to False if the image is already cropped
            and centered.

    Returns:
        polar_image (Array): 2D image in polar coordinates (sector-shaped scan).
    """
    if find_scan_cone:
        cartesian_image = fit_and_crop_around_scan_cone(cartesian_image)
    polar_image = cartesian_to_polar_matrix(
        cartesian_image,
        fill_value=fill_value,
        angle=angle,
        polar_shape=output_size,
        interpolation_order=interpolation_order,
    )
    return polar_image


def frustum_convert_rtp2xyz(rho, theta, phi):
    """Convert coordinates from (rho, theta, phi) space to (X,Y,Z) space using
    the frustum coordinate conversion.

    Angles are defined in radians.

    Args:
        rho (ndarray): Radial coordinates of the points to convert.
        theta (ndarray): Theta coordinates of the points to convert.
        phi (ndarray): Phi coordinates of the points to convert.

    Returns:
        x (ndarray): X coordinates of the converted points.
        y (ndarray): Y coordinates of the converted points.
        z (ndarray): Z coordinates of the converted points.
    """
    if ops.size(rho) != ops.size(theta) or ops.size(rho) != ops.size(phi):
        raise ValueError("Number of elements in rho, theta, and phi should be the same")

    z = rho / ops.sqrt(1 + ops.tan(theta) ** 2 + ops.tan(phi) ** 2)
    x = z * ops.tan(theta)
    y = z * ops.tan(phi)

    return x, y, z


def frustum_convert_rt2xz(rho, theta):
    """Convert coordinates from (rho, theta) space to (X,Z) space using
    the frustum coordinate conversion.

    Angles are defined in radians.

    Args:
        rho (ndarray): Radial coordinates of the points to convert.
        theta (ndarray): Theta coordinates of the points to convert.

    Returns:
        x (ndarray): X coordinates of the converted points.
        z (ndarray): Z coordinates of the converted points.
    """
    if ops.size(rho) != ops.size(theta):
        raise ValueError("Number of elements in rho and theta should be the same")

    z = rho / ops.sqrt(1 + ops.tan(theta) ** 2)
    x = z * ops.tan(theta)

    return x, z


def frustum_convert_xz2rt(x, z, theta_limits):
    """Convert coordinates from (X,Z) space to (rho, theta) space using
    the frustum coordinate conversion.

    Angles are defined in radians.

    Args:
        x (ndarray): X coordinates of the points to convert.
        z (ndarray): Z coordinates of the points to convert.
        theta_limits (list): Theta limits of the original volume. Any
            point that resides outside of these limits is potentially
            undefined, and therefore, the radial value for these points is
            made to be -1.

    Returns:
        rho (ndarray): Radial coordinates of the converted points.
        theta (ndarray): Theta coordinates of the converted points.
    """
    if ops.size(x) != ops.size(z):
        raise ValueError("Number of elements in x and z should be the same")

    rho = ops.sqrt(x**2 + z**2)
    theta = ops.arctan2(x, z)

    rho = ops.where(
        (rho < 0) | (theta < theta_limits[0]) | (theta > theta_limits[1]),
        -1,
        rho,
    )

    return rho, theta


def frustum_convert_xyz2rtp(x, y, z, theta_limits, phi_limits):
    """Convert coordinates from (X,Y,Z) space to (rho, theta, phi) space using
    the frustum coordinate conversion.

    Angles are defined in radians.

    Args:
        x (ndarray): X coordinates of the points to convert.
        y (ndarray): Y coordinates of the points to convert.
        z (ndarray): Z coordinates of the points to convert.
        tlimits, plimits:
            Theta and phi limits, respectively, of the original volume. Any
            point that resides outside of these limits is potentially
            undefined, and therefore, the radial value for these points is
            made to be -1.

    Returns:
        rho (ndarray): Radial coordinates of the converted points.
        theta (ndarray): Theta coordinates of the converted points.
        phi (ndarray): Phi coordinates of the converted points.
    """
    if ops.size(x) != ops.size(y) or ops.size(x) != ops.size(z):
        raise ValueError("Number of elements in x, y, and z should be the same")

    rho = ops.sqrt(x**2 + y**2 + z**2)
    theta = ops.arctan2(x, z)
    phi = ops.arctan2(y, z)

    rho = ops.where(
        (rho < 0)
        | (theta < theta_limits[0])
        | (theta > theta_limits[1])
        | (phi < phi_limits[0])
        | (phi > phi_limits[1]),
        -1,
        rho,
    )

    return rho, theta, phi
